from tensorflow.python.keras import Model
from tensorflow.python.keras.layers import Layer
from tensorflow.python.util import nest

__all__ = [
    'to_keras_objective', 'copy_keras_metadata', 'has_keras_meta',
    'add_trainable_weights'
]


def has_keras_meta(tensor):
  return hasattr(tensor, '_keras_history') and hasattr(tensor, '_keras_mask')


def copy_keras_metadata(keras_tensor, *new_tensors):
  if not hasattr(keras_tensor, '_keras_history') or \
  not hasattr(keras_tensor, '_keras_mask'):
    pass
  else:
    new_tensors = nest.flatten(new_tensors)
    history = keras_tensor._keras_history
    mask = keras_tensor._keras_mask
    for t in new_tensors:
      setattr(t, '_keras_history', history)
      setattr(t, '_keras_mask', mask)
  return new_tensors[0] if len(new_tensors) == 1 else new_tensors


def add_trainable_weights(layer, *variables):
  from odin.backend import is_variable
  variables = nest.flatten(variables)
  assert all(is_variable(v) for v in variables), \
  "All objects from variables must be instance of tensorflow.Variable"

  assert isinstance(layer, Layer), \
  "layer must be instance of tensorflow.python.keras.layers.Layer"

  variables = [v for v in variables if v not in layer._trainable_weights]
  layer._trainable_weights = layer._trainable_weights + variables
  return layer



# ===========================================================================
# Using TIDIGITS dataset to predict gender (Boy, Girl, Woman, Man)
# ===========================================================================
# Saved WAV file format:
#     0) [train|test]
#     1) [m|w|b|g] (alias for man, women, boy, girl)
#     2) [age]
#     3) [dialectID]
#     4) [speakerID]
#     5) [production]
#     6) [digit_sequence]
#     => "train_g_08_17_as_a_4291815"
# ===========================================================================
from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'gpu,float32,seed=1234'
import shutil
import pickle

import numpy as np
import tensorflow as tf

from sklearn.metrics import confusion_matrix, accuracy_score

from odin import backend as K, nnet as N, fuel as F, visual as V
from odin.stats import train_valid_test_split, freqcount
from odin import ml
from odin import training
from odin import preprocessing as pp
from odin.visual import print_dist, print_confusion, print_hist
from odin.utils import (Progbar, unique_labels, chain, get_formatted_datetime,
                        as_tuple_of_shape, stdio, ctext, ArgController)

from utils import prepare_data, get_exp_path
# ===========================================================================
# Input arguments
# ===========================================================================
args = ArgController(
).add('-nmix', "Number of GMM mixture", 256
).add('-tdim', "Dimension of t-matrix", 128
).add('-feat', "Acoustic feature: spec, mspec, mfcc, bnf, sdc", 'bnf'
).add('-task', 'gender, age, dialect, speaker, digit', 'gender'
).add('--retrain', "deleted trained model, and re-train everything", False
).parse()
# ===========================================================================
# Const
# ===========================================================================
DTYPE = 'float64'
# ====== GMM trainign ====== #
NMIX = args.nmix
GMM_NITER = 10
GMM_DOWNSAMPLE = 4
GMM_STOCHASTIC = True
# ====== IVEC training ====== #
TV_DIM = args.tdim
TV_NITER = 10
# ===========================================================================
# path and dataset
# ===========================================================================
EXP_DIR, MODEL_PATH, LOG_PATH = get_exp_path('ivec', args, override=args.retrain)
stdio(LOG_PATH)
X, train, y_train, test, y_test, labels = prepare_data(
    feat=args.feat, label=args.task, for_ivec=True)
# ===========================================================================
# Training Ivector
# ===========================================================================
ivec = ml.Ivector(path=EXP_DIR, nmix=NMIX, tv_dim=TV_DIM,
                  nmix_start=1, niter_gmm=GMM_NITER, niter_tmat=TV_NITER,
                  allow_rollback=True, exit_on_error=False,
                  downsample=GMM_DOWNSAMPLE, stochastic_downsample=GMM_STOCHASTIC,
                  device='gpu', ncpu=1, gpu_factor_gmm=80, gpu_factor_tmat=3,
                  dtype=DTYPE, seed=1234, name=None)
ivec.fit(X=X, indices=train,
         extract_ivecs=True, keep_stats=True)
ivec.transform(X=X, indices=test,
               save_ivecs=True, keep_stats=True, name='test')
# ====== i-vector ====== #
I_train = F.MmapData(path=ivec.get_i_path(None), read_only=True)
F_train = F.MmapData(path=ivec.get_f_path(None), read_only=True)
name_train = np.genfromtxt(ivec.get_name_path(None), dtype=str)

I_test = F.MmapData(path=ivec.get_i_path(name='test'), read_only=True)
F_test = F.MmapData(path=ivec.get_f_path(name='test'), read_only=True)
name_test = np.genfromtxt(ivec.get_name_path(name='test'), dtype=str)
# ===========================================================================
# I-vector
# ===========================================================================
X_train = I_train
X_test = I_test
# ====== cosine scoring ====== #
print(ctext("==== '%s'" % "Ivec cosine-scoring", 'cyan'))
scorer = ml.Scorer(centering=True, wccn=True, lda=True, method='cosine')
scorer.fit(X=X_train, y=y_train)
scorer.evaluate(X_test, y_test, labels=labels)
# ====== plda scoring ====== #
print(ctext("==== '%s'" % "Ivec PLDA-scoring", 'cyan'))
scorer = ml.PLDA(n_phi=100, n_iter=12,
                 centering=True, wccn=True, unit_length=True,
                 random_state=1234)
scorer.fit(X=X_train, y=y_train)
scorer.evaluate(X_test, y_test, labels=labels)
# ====== svm scoring ====== #
print(ctext("==== '%s'" % "Ivec SVM-scoring", 'cyan'))
scorer = ml.Scorer(wccn=True, lda=True, method='svm')
scorer.fit(X=X_train, y=y_train)
scorer.evaluate(X_test, y_test, labels=labels)



__version__ = "1.0.0"

import os

# this should always be true, the gain in performance
# for preempting whole GPU memory is marginal (memory fragmentation)
# it further prevent you from running multiple experiments
# on 1 GPU, take all memory from other processes even though
# it does not use all the computational resources
os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'



from __future__ import print_function, division, absolute_import

import os
import marshal
from six import string_types
from collections import OrderedDict
from contextlib import contextmanager
from typing import Iterable, Text, Tuple, Union, List

import numpy as np

__all__ = [
  'get_total_opened_mmap',
  'read_mmaparray_header',
  'MmapArrayWriter',
  'MmapArray',
]

MAX_OPEN_MMAP = 120
_INSTANCES_WRITER = OrderedDict()
_HEADER = b'mmapdata'
_MAXIMUM_HEADER_SIZE = 486

# ===========================================================================
# Helper
# ===========================================================================
def get_total_opened_mmap():
  return len(_INSTANCES_WRITER)

def read_mmaparray_header(path):
  """
  Parameters
  ----------

  Return
  ------
  dtype, shape
    Necessary information to create numpy.memmap
  """
  with open(path, mode='rb') as f:
    # ====== check header signature ====== #
    try:
      if f.read(len(_HEADER)) != _HEADER:
        raise Exception
    except Exception as e:
      raise Exception('Invalid header for MmapData.')
    # ====== 8 bytes for size of info ====== #
    try:
      size = int(f.read(8))
      dtype, shape = marshal.loads(f.read(size))
    except Exception as e:
      f.close()
      raise Exception('Error reading memmap data file: %s' % str(e))
    # ====== return file object ====== #
    return dtype, shape

def _aligned_memmap_offset(dtype):
  header_size = len(_HEADER) + 8 + _MAXIMUM_HEADER_SIZE
  type_size = np.dtype(dtype).itemsize
  n = np.ceil(header_size / type_size)
  return int(n * type_size)

# ===========================================================================
# Writing new memory-mapped array
# ===========================================================================
class MmapArrayWriter(object):
  """
  Parameters
  ----------
  path : str
    path to a file for writing memory-mapped data

  shape : tuple
    pass

  dtype : numpy.dtype
    data type

  remove_exist : boolean (default=False)
    if file at given path exists, remove it
  """
  def __new__(cls, path, *args, **kwargs):
    path = os.path.abspath(path)
    # Found old instance
    if path in _INSTANCES_WRITER:
      obj = _INSTANCES_WRITER[path]
      if not obj.is_closed:
        return obj
    # ====== increase memmap count ====== #
    if get_total_opened_mmap() > MAX_OPEN_MMAP:
      raise RuntimeError(
        "Only allowed to open maximum of %d memmap file" % MAX_OPEN_MMAP)
    # ====== create new instance ====== #
    new_instance = super(MmapArrayWriter, cls).__new__(cls)
    _INSTANCES_WRITER[path] = new_instance
    return new_instance

  def __init__(self, path: Text, shape: Tuple,
               dtype: Union[Text, np.dtype] = 'float32',
               remove_exist: bool = False):
    super(MmapArrayWriter, self).__init__()
    assert isinstance(path, string_types), "path must be string or text."
    # validate path
    path = os.path.abspath(path)
    # ====== remove exist ====== #
    if remove_exist and os.path.exists(path):
      if os.path.isfile(path):
        os.remove(path)
      else:
        raise RuntimeError(
          "Give path at '%s' is a folder, cannot remove!" % path)
    # ====== check shape info ====== #
    if not isinstance(shape, Iterable):
      shape = (shape,)
    shape = tuple([0 if i is None or i < 0 else int(i)
                   for i in shape])
    # ====== read exist file ====== #
    if os.path.exists(path):
      dtype, shape = read_mmaparray_header(path)
      f = open(path, 'rb+')
    # ====== create new file ====== #
    else:
      if dtype is None or shape is None:
        raise Exception("First created this MmapData, `dtype` and "
                        "`shape` must NOT be None.")
      f = open(path, 'wb+')
      f.write(_HEADER)
      dtype = str(np.dtype(dtype))
      if isinstance(shape, np.ndarray):
        shape = shape.tolist()
      if not isinstance(shape, (tuple, list)):
        shape = (shape,)
      _ = marshal.dumps([dtype, shape])
      size = len(_)
      if size > _MAXIMUM_HEADER_SIZE:
        raise Exception('The size of header excess maximum allowed size '
                        '(%d bytes).' % _MAXIMUM_HEADER_SIZE)
      size = '%8d' % size
      f.write(size.encode())
      f.write(_)
    # ====== assign attributes ====== #
    self._file = f
    self._path = path
    data = np.memmap(f, dtype=dtype, shape=shape,
                     mode='r+', offset=_aligned_memmap_offset(dtype))
    self._data = data
    self._is_closed = False

  @property
  def shape(self):
    return self._data.shape

  @property
  def path(self):
    return self._path

  @property
  def is_closed(self):
    return self._is_closed

  def flush(self):
    self._data.flush()

  def close(self):
    if self.is_closed:
      return
    self._is_closed = True
    if self.path in _INSTANCES_WRITER:
      del _INSTANCES_WRITER[self.path]
    # flush in read-write mode
    self.flush()
    # close mmap and file
    self._data._mmap.close()
    del self._data
    self._file.close()

  def _resize(self, new_length):
    # ====== local files ====== #
    f = self._file
    old_length = self._data.shape[0]
    # ====== check new shape ====== #
    if new_length < old_length:
      raise ValueError(
        'Only support extending memmap, and cannot shrink the memory.')
    # nothing to resize
    elif new_length == old_length:
      return self
    # ====== flush previous changes ====== #
    # resize by create new memmap and also rename old file
    shape = (new_length,) + self._data.shape[1:]
    dtype = self._data.dtype.name
    # rewrite the header, update metadata
    f.seek(len(_HEADER))
    meta = marshal.dumps([dtype, shape])
    size = '%8d' % len(meta)
    f.write(size.encode(encoding='utf-8'))
    f.write(meta)
    f.flush()
    # close old data
    self._data._mmap.close()
    del self._data
    # extend the memmap, by open numpy.memmap with bigger shape
    f = open(self.path, 'rb+')
    self._file = f
    mmap = np.memmap(f, dtype=dtype, shape=shape,
                     mode='r+', offset=_aligned_memmap_offset(dtype))
    self._data = mmap
    return self

  def write(self, arrays: Iterable):
    """ Extending the memory-mapped data and copy the array
    into extended area. """
    if self.is_closed:
      raise RuntimeError("The MmapArrayWriter is closed!")
    # only get arrays matched the shape
    add_size = 0
    if not isinstance(arrays, Iterable) or isinstance(arrays, np.ndarray):
      arrays = (arrays,)
    accepted_arrays = []
    # ====== check if shape[1:] matching ====== #
    for a in arrays:
      if a.shape[1:] == self._data.shape[1:]:
        accepted_arrays.append(a)
        add_size += a.shape[0]
    # no new array to append
    if len(accepted_arrays) == 0:
      raise RuntimeError(
        "No appropriate array found for writing, given: %s, "
        "; but require array with shape: %s" %
        (','.join([str(i.shape) for i in arrays]), self.shape))
    # ====== resize ====== #
    old_size = self._data.shape[0]
    # special case, Mmap is init with temporary size = 1 (all zeros),
    # NOTE: risky to calculate sum of big array here
    if old_size == 1 and \
    sum(np.sum(np.abs(d[:])) for d in self._data) == 0.:
      old_size = 0
    # resize and append data
    # resize only once will be faster
    self._resize(old_size + add_size)
    # ====== update values ====== #
    data = self._data
    for a in accepted_arrays:
      data[old_size:old_size + a.shape[0]] = a
      old_size += a.shape[0]
    return self

  def __enter__(self):
    return self

  def __exit__(self, *exc):
    self.close()

  def __del__(self):
    self.close()

# ===========================================================================
# Reading the array
# ===========================================================================
class MmapArray(np.memmap):
  """Create a memory-map to an array stored in a *binary* file on disk.

    Memory-mapped files are used for accessing small segments of large files
    on disk, without reading the entire file into memory.  NumPy's
    memmap's are array-like objects.  This differs from Python's ``mmap``
    module, which uses file-like objects.

    This subclass of ndarray has some unpleasant interactions with
    some operations, because it doesn't quite fit properly as a subclass.
    An alternative to using this subclass is to create the ``mmap``
    object yourself, then create an ndarray with ndarray.__new__ directly,
    passing the object created in its 'buffer=' parameter.

    This class may at some point be turned into a factory function
    which returns a view into an mmap buffer.

    Delete the memmap instance to close the memmap file.

    Parameters
    ----------
    path : str, file-like object, or pathlib.Path instance
        The file name or file object to be used as the array data buffer.
    mode : {'r+', 'r', 'w+', 'c'}, optional
        The file is opened in this mode:

        +------+-------------------------------------------------------------+
        | 'r'  | Open existing file for reading only.                        |
        +------+-------------------------------------------------------------+
        | 'r+' | Open existing file for reading and writing.                 |
        +------+-------------------------------------------------------------+
        | 'w+' | Create or overwrite existing file for reading and writing.  |
        +------+-------------------------------------------------------------+
        | 'c'  | Copy-on-write: assignments affect data in memory, but       |
        |      | changes are not saved to disk.  The file on disk is         |
        |      | read-only.                                                  |
        +------+-------------------------------------------------------------+
        Default is 'r+'.
  """
  def __new__(subtype, path, mode='r+'):
    path = os.path.abspath(path)
    if not os.path.exists(path) and os.path.isfile(path):
      raise ValueError('path must be existed file created by MmapArrayWriter.')
    dtype, shape = read_mmaparray_header(path)
    offset = _aligned_memmap_offset(dtype)
    new_array = super(MmapArray, subtype).__new__(
      subtype=subtype, filename=path, dtype=dtype, mode=mode,
      offset=offset, shape=shape)
    new_array._path = path
    return new_array

  @property
  def path(self):
    return self._path



from __future__ import print_function, division, absolute_import

import numpy as np

def read(path, boxes=None):
  """
  Return
  ------
  Always return 3D images
  (n_frames, channels, width, height)
  """
  import imageio
  vid = imageio.get_reader(path)
  metadata = vid.get_meta_data()
  fps = metadata['fps']
  nb_frames = metadata['nframes']
  if boxes is not None:
    pass
  print(nb_frames)
  exit()
  try:
    frames = []
    for i in vid:
      # it is bizzare why width and height are swapped
      if i.ndim == 3: # swap channel first
        i = i.transpose(2, 1, 0)
      else:
        i = np.expand_dims(i.transpose(1, 0), 1)
      frames.append(i)
  except RuntimeError:
    pass
  frames = np.array(frames, dtype=frames[0].dtype)
  return frames, fps



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'float32,gpu'
import sys
import time
from collections import defaultdict

import numpy as np

from odin import visual as V, nnet as N
from odin.utils import (ctext, unique_labels, UnitTimer,
                        Progbar, get_logpath, get_module_from_path,
                        get_script_path, mpi, catch_warnings_ignore)
from odin import fuel as F, preprocessing as pp
from odin.stats import sampling_iter

from helpers import (PATH_ACOUSTIC_FEATURES, EXP_DIR,
                     ALL_FILES, IS_DEBUGGING, FEATURE_RECIPE, FEATURE_NAME,
                     ALL_DATASET, Config, NCPU,
                     validate_features_dataset)
# ALL_FILES
# Header:
#  0       1      2      3       4          5         6
# path, channel, name, spkid, dataset, start_time, end_time
# ===========================================================================
# Extractor
# ===========================================================================
recipe = get_module_from_path(identifier=FEATURE_RECIPE,
                              path=get_script_path(),
                              prefix='feature_recipes')
if len(recipe) == 0:
  raise ValueError("Cannot find feature recipe with name: '%s'" % FEATURE_RECIPE)
recipe = recipe[0]()
# ===========================================================================
# Debug mode
# ===========================================================================
if IS_DEBUGGING:
  with np.warnings.catch_warnings():
    rand = np.random.RandomState(seed=Config.SUPER_SEED)
    np.warnings.filterwarnings('ignore')
    # ====== stratify sampling from each dataset ====== #
    clusters = defaultdict(list)
    clusters_count = defaultdict(int)
    samples = []
    for row in sorted(ALL_FILES,
                      key=lambda x: x[0]):
      clusters[row[4]].append(row)
      clusters_count[row[4]] += 1
    for k, v in clusters.items():
      rand.shuffle(v)
      samples += v[:18] # 18 files from each dataset
    # ====== run the MPI for feature extraction ====== #
    prog = Progbar(target=len(samples),
                   print_report=True, print_summary=False,
                   name=FEATURE_RECIPE)
    error_signal = []
    for feat in mpi.MPI(jobs=samples,
                        func=recipe.transform,
                        ncpu=NCPU, batch=1):
      assert FEATURE_NAME in feat
      # update progress
      if isinstance(feat, pp.base.ExtractorSignal):
        error_signal.append(feat)
        prog.add(1)
        continue
      prog['spkid'] = feat['spkid']
      prog['name'] = feat['name']
      prog['dsname'] = feat['dsname']
      prog['duration'] = feat['duration']
      prog.add(1)
      # 30% chance plotting
      if rand.rand() < 0.5:
        V.plot_multiple_features(feat, fig_width=20,
                                 title='[%s]%s' % (feat['dsname'], feat['name']))
    V.plot_save(os.path.join(EXP_DIR, 'debug_%s.pdf' % FEATURE_RECIPE),
                dpi=30)
    # ====== save the extractor debugging log ====== #
    pp.set_extractor_debug(recipe, debug=True)
    recipe.transform(samples[0])
    with open(os.path.join(EXP_DIR, 'debug_%s.log' % FEATURE_RECIPE), 'w') as f:
      for name, step in recipe.steps:
        f.write(step.last_debugging_text)
      # ====== print error signal ====== #
      for e in error_signal:
        f.write(str(e) + '\n')
        print(e)
  exit()
# ===========================================================================
# Running the extractor
# ===========================================================================
# ====== basic path ====== #
output_dataset_path = os.path.join(PATH_ACOUSTIC_FEATURES, FEATURE_RECIPE)

processor_log_path = os.path.join(EXP_DIR, 'processor_%s.log' % FEATURE_RECIPE)
if os.path.exists(processor_log_path):
  os.remove(processor_log_path)
print("Log path:", ctext(processor_log_path, 'cyan'))

ds_validation_path = os.path.join(EXP_DIR, 'validate_%s.pdf' % FEATURE_RECIPE)
if os.path.exists(ds_validation_path):
  os.remove(ds_validation_path)
print("Validation path:", ctext(ds_validation_path, 'cyan'))

# ====== running the processing ====== #
with catch_warnings_ignore(Warning):
  processor = pp.FeatureProcessor(
      jobs=ALL_FILES,
      path=output_dataset_path,
      extractor=recipe,
      n_cache=320,
      ncpu=NCPU,
      override=True,
      identifier='name',
      log_path=processor_log_path,
      stop_on_failure=False)
  processor.run()
# ===========================================================================
# Make some visualization
# ===========================================================================
validate_features_dataset(output_dataset_path, ds_validation_path)



from __future__ import absolute_import, division, print_function

import os
from collections import defaultdict
from typing import Dict, Text

_FIGURE_LIST = defaultdict(dict)


class Visualizer(object):
  """ Visualizer """

  def assert_figure(self, fig):
    from matplotlib import pyplot as plt
    assert isinstance(fig, plt.Figure), \
    'fig must be instance of matplotlib.Figure, but given: %s' % str(type(fig))
    return fig

  def assert_axis(self, ax):
    from matplotlib import pyplot as plt
    from odin.visual.figures import to_axis
    ax = to_axis(ax)
    assert isinstance(ax, plt.Axes), \
    'ax must be instance of matplotlib.Axes, but given: %s' % str(type(ax))
    return ax

  def get_figures(self) -> Dict[Text, 'Figure']:
    return _FIGURE_LIST[id(self)]

  def add_figure(self, name, fig):
    from matplotlib import pyplot as plt
    assert isinstance(fig, plt.Figure), \
    'fig must be instance of matplotlib.Figure'
    _FIGURE_LIST[id(self)][name] = fig
    return self

  def save_figures(self,
                   path='/tmp/tmp.pdf',
                   dpi=None,
                   separate_files=True,
                   clear_figures=True,
                   verbose=False):
    from odin.utils import ctext
    from matplotlib import pyplot as plt
    # checking arguments
    if os.path.isfile(path) or '.pdf' == path[-4:].lower():
      separate_files = False
      assert '.pdf' == path[-4:].lower(), \
      "If a file is given, it must be PDF file"
    figures = _FIGURE_LIST[id(self)]
    n_figures = len(figures)
    if n_figures == 0:
      return self
    # ====== saving PDF file ====== #
    if verbose:
      print("Saving %s figures to path: " % ctext(n_figures, 'lightcyan'),
            ctext(path, 'lightyellow'))

    if not separate_files:
      if dpi is None:
        dpi = 48
      if '.pdf' not in path:
        path = path + '.pdf'
      from matplotlib.backends.backend_pdf import PdfPages
      pp = PdfPages(path)
      for key, fig in figures.items():
        try:
          fig.savefig(pp, dpi=dpi, format='pdf', bbox_inches="tight")
          if verbose:
            print(" - Saved '%s' to pdf file" % ctext(key, 'cyan'))
        except Exception as e:
          if verbose:
            print(" - Error '%s'" % ctext(key, 'cyan'))
            print("  ", e)
      pp.close()
    # ====== saving PNG file ====== #
    else:
      if dpi is None:
        dpi = 160
      if not os.path.exists(path):
        os.mkdir(path)
      assert os.path.isdir(path), "'%s' must be path to a folder" % path
      kwargs = dict(dpi=dpi, bbox_inches="tight")
      for key, fig in figures.items():
        out_path = os.path.join(path, key + '.png')
        try:
          fig.savefig(out_path, **kwargs)
          if verbose:
            print(" - Saved '%s' to %s" %
                  (ctext(key, 'cyan'), ctext(out_path, 'yellow')))
        except Exception as e:
          if verbose:
            print(" - Error '%s'" % ctext(key, 'cyan'))
            print("  ", e)
    # ====== clear figures ====== #
    if clear_figures:
      for fig in figures.values():
        plt.close(fig)
      figures.clear()
    return self



from __future__ import absolute_import, division, print_function

import inspect
import os
from collections import Mapping
from contextlib import contextmanager

from six.moves import builtins, cPickle

from odin.backend import (keras_callbacks, keras_helpers, losses, metrics,
                          tf_utils)
from odin.backend.tensor import *
from odin.utils import as_tuple, is_path, is_string



from __future__ import print_function, division, absolute_import

import os
import io
import mmap
import marshal
import sqlite3
from itertools import chain
from six import add_metaclass
from six.moves import cPickle
from contextlib import contextmanager
from abc import ABCMeta, abstractmethod, abstractproperty
from collections import OrderedDict, Iterator, defaultdict, MutableMapping, Mapping

import numpy as np

from odin.utils import async, is_string, ctext

__all__ = [
  'NoSQL',
  'MmapDict',
  'TableDict',
  'SQLiteDict'
]

# ===========================================================================
# Helpers
# ===========================================================================
def _dump(x):
  if isinstance(x, np.str_):
    x = str(x)
  return marshal.dumps(x)

@add_metaclass(ABCMeta)
class NoSQL(MutableMapping):
  """ The idea of this is transform everything (even SQL) into
  (key, value) kind-of database.
  """
  _INSTANCES = defaultdict(dict)

  def __new__(subclass, path, read_only=False, cache_size=250,
              *args, **kwargs):
    if not is_string(path):
      raise ValueError("`path` for MmapDict must be string, but given "
                       "object with type: %s" % type(path))
    path = os.path.abspath(path)
    read_only = bool(read_only)
    cache_size = int(cache_size)
    # get stored instances
    all_instances = NoSQL._INSTANCES[subclass.__name__]
    # ====== Found pre-defined instance ====== #
    if path in all_instances:
      return all_instances[path]
    # ====== Create new instance ====== #
    new_instance = super(NoSQL, subclass).__new__(subclass)
    all_instances[path] = new_instance
    # some pre-defined attribute
    new_instance._cache_size = cache_size
    new_instance._read_only = read_only
    new_instance._new_args_called = False
    new_instance._path = path
    new_instance._is_closed = False
    return new_instance

  def __init__(self, path, read_only=False, cache_size=250,
               override=False):
    super(NoSQL, self).__init__()
    # ====== check override ====== #
    if override and os.path.exists(path) and os.path.isfile(path):
      os.remove(path)
    # ====== init ====== #
    try:
      self._restore_dict(self.path, self.read_only, self.cache_size)
    except Exception as e:
      # exception when restoring the data,
      # remove singleton instance first
      del NoSQL._INSTANCES[self.__class__.__name__][self.path]
      import traceback
      traceback.print_exc()
      raise e

  # ==================== abstract methods ==================== #
  @abstractmethod
  def _restore_dict(self, path, read_only, cache_size):
    raise NotImplementedError

  @abstractmethod
  def _close(self):
    raise NotImplementedError

  @abstractmethod
  def _flush(self, save_all):
    raise NotImplementedError

  # ==================== pickling ==================== #
  def __getstate__(self):
    if not self._new_args_called:
      raise RuntimeError(
          "You must use argument `protocol=cPickle.HIGHEST_PROTOCOL` "
          "when using `pickle` or `cPickle` to be able pickling NoSQL.")
    self._new_args_called = False
    return self.path, self.read_only, self.cache_size

  def __setstate__(self, states):
    path, read_only, cache_size = states
    if not os.path.exists(path):
      raise ValueError("Cannot find store NoSQL database at path: %s."
                       "If you have moved the database, the dumps from "
                       "cannot restore the previous intance." % path)
    self._restore_dict(path, read_only, cache_size)
    self._path = path
    self._read_only = read_only
    self._cache_size = cache_size

  def __getnewargs__(self):
    self._new_args_called = True
    return (self.path,)

  # ==================== Abstract properties ==================== #
  @property
  def cache_size(self):
    return self._cache_size

  @property
  def read_only(self):
    return self._read_only

  @property
  def path(self):
    return self._path

  @property
  def is_closed(self):
    return self._is_closed

  def close(self):
    if self._is_closed:
      return
    self._is_closed = True
    # check if in read only mode
    if not self.read_only:
      self.flush(save_all=True)
    # delete Singleton instance
    del NoSQL._INSTANCES[self.__class__.__name__][self.path]
    # close but some of the attribute may not be initialized
    try:
      self._close()
    except AttributeError as e:
      pass

  def flush(self, save_all=True):
    if self.read_only or self.is_closed:
      return
    self._flush(save_all=bool(save_all))
    return self

  # ==================== dictionary methods ==================== #
  def __call__(self, key):
    return self.__getitem__(key)

  def set(self, key, value):
    self.__setitem__(key, value)
    return self

  @abstractmethod
  def __setitem__(self, key, value):
    raise NotImplementedError

  @abstractmethod
  def __getitem__(self, key):
    raise NotImplementedError

  @abstractmethod
  def __delitem__(self, key):
    raise NotImplementedError

  @abstractmethod
  def __len__(self):
    raise NotImplementedError

  def __iter__(self):
    return self.items()

  @abstractmethod
  def keys(self):
    raise NotImplementedError

  @abstractmethod
  def values(self):
    raise NotImplementedError

  @abstractmethod
  def items(self):
    raise NotImplementedError

  def __del__(self):
    self.close()

  def __repr__(self):
    return str(self)

  def update(*args, **kwargs):
    self = args[0]
    if self.read_only:
      return self
    return super(NoSQL, self).update(*args, **kwargs)

  def clear(self):
    if self.read_only:
      return self
    return super(NoSQL, self).clear()

  def __cmp__(self, d):
    if isinstance(d, NoSQL):
      return self.path == d.path
    raise ValueError("Cannot compare NoSQL to other type (i.e. %s)"
                     % d.__class__.__name__)


# ===========================================================================
# MmapDict
# ===========================================================================
def _safe_loading_indices(file_obj, class_name, path):
  try:
    return cPickle.loads(file_obj)
  except Exception as e:
    del NoSQL._INSTANCES[class_name][path]
    import traceback
    traceback.print_exc()
    raise e

class MmapDict(NoSQL):
  """ MmapDict
  Handle enormous dictionary (up to thousand terabytes of data) in
  memory mapped dictionary, extremely fast to load, and for randomly access.
  The alignment of saved files:

  ==> |'mmapdict'|48-bytes(max_pos)|48-bytes(dict_size)|MmapData|indices-dict|

  * The first 48-bytes number: is ending position of the Mmmap, start from
  the (8 + 48 + 48) bytes.

  * The next 48-bytes number: is length of pickled indices (i.e. the
  indices start from max_pos to max_pos + dict_size)

  Note
  ----
  Only support (key, value) types = (str, primitive_type)
  MmapDict read speed is double faster than SQLiteDict.
  MmapDict also support multiprocessing
  """
  HEADER = b'mmapdict'
  SIZE_BYTES = 48
  # the indices are flushed after it is increased this amount of size
  MAX_INDICES_SIZE = 25 # in megabyte

  def _restore_dict(self, path, read_only, cache_size):
    # ====== already exist ====== #
    if os.path.exists(path):
      if os.path.getsize(path) == 0:
        if read_only:
          raise Exception('File at path:"%s" has zero size, no data '
                          'found in (read-only mode).' % path)
      file = open(str(path), mode='rb+')
      if file.read(len(MmapDict.HEADER)) != MmapDict.HEADER:
        file.close() # close the file before Exception
        raise Exception('Given file is not in the right format '
                        'for MmapDict.')
      # 48 bytes for the file size
      max_position = int(file.read(MmapDict.SIZE_BYTES))
      # length of pickled indices dictionary
      dict_size = int(file.read(MmapDict.SIZE_BYTES))
      # read dictionary
      file.seek(max_position)
      pickled_indices = file.read(dict_size)
      self._indices_dict = async(_safe_loading_indices)(
          pickled_indices, self.__class__.__name__, path)
    # ====== create new file from scratch ====== #
    else:
      file = open(str(path), mode='wb+')
      file.write(MmapDict.HEADER)
      # just write the header
      header = ('%' + str(MmapDict.SIZE_BYTES) + 'd') % \
          (len(MmapDict.HEADER) + MmapDict.SIZE_BYTES * 2)
      file.write(header.encode())
      # write the length of Pickled indices dictionary
      data_size = ('%' + str(MmapDict.SIZE_BYTES) + 'd') % 0
      file.write(data_size.encode())
      file.flush()
      # init indices dict
      self._indices_dict = {}
    # ====== create Mmap from offset file ====== #
    self._file = file
    self._mmap = mmap.mmap(file.fileno(), length=0, offset=0,
                           flags=mmap.MAP_SHARED)
    self._increased_indices_size = 0. # in MB
    # store all the (key, value) recently added
    self._cache_dict = {}

  def _close(self):
    self._mmap.close()
    self._file.close()
    del self._indices_dict
    del self._cache_dict

  def _flush(self, save_all=False):
    """
    Parameters
    ----------
    save_all: bool
        force the indices dictionary to be saved, even though,
        its increased hasn't reach the maximum.
    """
    # check if closed or in read only mode
    if self.is_closed or self.read_only:
      return
    # ====== write new data ====== #
    # get old position
    file = self._file
    # start from header (i.e. "mmapdict")
    file.seek(len(MmapDict.HEADER))
    max_position = int(file.read(MmapDict.SIZE_BYTES))
    # ====== serialize the data ====== #
    # start from old_max_position, append new values
    file.seek(max_position)
    for key, value in self._cache_dict.items():
      try:
        value = _dump(value)
      except ValueError:
        raise RuntimeError("Cannot marshal.dump %s" % str(value))
      self.indices[key] = (max_position, len(value))
      max_position += len(value)
      file.write(value)
      # increase indices size (in MegaBytes)
      self._increased_indices_size += (8 + 8 + len(key)) / 1024. / 1024.
    # ====== write the dumped indices ====== #
    indices_length = 0
    if save_all or \
    self._increased_indices_size > MmapDict.MAX_INDICES_SIZE:
      indices_dump = cPickle.dumps(self.indices,
                                   protocol=cPickle.HIGHEST_PROTOCOL)
      indices_length = len(indices_dump)
      file.write(indices_dump)
      self._increased_indices_size = 0.
    # ====== update the position ====== #
    # write new max size
    file.seek(len(MmapDict.HEADER))
    max_position = ('%' + str(MmapDict.SIZE_BYTES) + 'd') % max_position
    file.write(max_position.encode())
    # update length of pickled indices dictionary
    if indices_length > 0:
      indices_length = ('%' + str(MmapDict.SIZE_BYTES) + 'd') % indices_length
      file.write(indices_length.encode())
    # flush everything
    file.flush()
    # upate the mmap
    self._mmap.close(); del self._mmap
    self._mmap = mmap.mmap(file.fileno(), length=0, offset=0,
                           flags=mmap.MAP_SHARED)
    # reset some values
    del self._cache_dict
    self._cache_dict = {}

  # ==================== I/O methods ==================== #
  @property
  def indices(self):
    if not isinstance(self._indices_dict, Mapping):
      self._indices_dict = self._indices_dict.get()
    return self._indices_dict

  @property
  def is_loaded(self):
    if self.is_closed:
      return False
    if not isinstance(self._indices_dict, Mapping):
      return self._indices_dict.finished
    return True

  def __str__(self):
    length = None if self.is_closed else \
        str(len(self.indices) + len(self._cache_dict))
    cache_length = 'None' if self.is_closed else \
        str(len(self._cache_dict))
    fmt = '<MmapDict path:"%s", length:%s/%s, loaded:%s, closed:%s, read_only:%s>'
    return fmt % (self.path, length, cache_length,
                  self.is_loaded, self.is_closed, self.read_only)

  # ==================== Dictionary ==================== #
  def __setitem__(self, key, value):
    if self.read_only:
      return
    key = str(key)
    # store newly added value for fast query
    self._cache_dict[key] = value
    if len(self._cache_dict) > self.cache_size:
      self.flush(save_all=False)

  def __getitem__(self, key):
    if key in self._cache_dict:
      return self._cache_dict[key]
    # ====== load from mmap ====== #
    start, size = self.indices[key]
    self._mmap.seek(start)
    return marshal.loads(self._mmap.read(size))

  def __contains__(self, key):
    return key in self.indices or key in self._cache_dict

  def __len__(self):
    return len(self.indices) + len(self._cache_dict)

  def __delitem__(self, key):
    if self.read_only:
      return
    if key in self._cache_dict:
      del self._cache_dict
    else:
      del self.indices[key]

  def keys(self):
    return chain(self.indices.keys(), self._cache_dict.keys())

  def values(self):
    for name, (start, size) in self.indices.items():
      self._mmap.seek(start)
      yield marshal.loads(self._mmap.read(size))
    for val in self._cache_dict.values():
      yield val

  def items(self):
    for name, (start, size) in self.indices.items():
      self._mmap.seek(start)
      yield name, marshal.loads(self._mmap.read(size))
    for key, val in self._cache_dict.values():
      yield key, val


# ===========================================================================
# SQLiteDict
# ===========================================================================
def _adapt_array(arr):
  """ Converts np.array to TEXT when inserting
  http://stackoverflow.com/a/31312102/190597 (SoulNibbler)
  """
  out = io.BytesIO()
  np.save(out, arr)
  out.seek(0)
  return sqlite3.Binary(out.read())


def _convert_array(text):
  """ Converts TEXT to np.array when selecting """
  out = io.BytesIO(text)
  out.seek(0)
  return np.load(out)


class TableDict(MutableMapping):

  def __init__(self, sqlite, table_name):
    if not isinstance(sqlite, SQLiteDict):
      raise ValueError("`sqlite` must be instance of SQLiteDict")
    self._sqlite = sqlite
    self._name = str(table_name)

  @contextmanager
  def table_context(self):
    """Return temporary context that switch the SQLite to given table"""
    curr_tab = self._sqlite.current_table
    self._sqlite.set_table(self._table_name)
    yield None
    self._sqlite.set_table(curr_tab)

  @property
  def sqlite(self):
    return self._sqlite

  @property
  def name(self):
    return self._name

  def flush(self):
    with self.table_context():
      self._sqlite.flush()

  @property
  def connection(self):
    return self._sqlite.connection

  @property
  def cursor(self):
    return self._sqlite.cursor

  @property
  def path(self):
    return self._sqlite.path

  @property
  def is_closed(self):
    return self._sqlite.is_closed

  # ==================== pickling and properties ==================== #
  def __setstate__(self, states):
    self._sqlite, self._name = states

  def __getstate__(self):
    return self.sqlite, self.name

  def __str__(self):
    return
    '<' + ctext('TableDict: ', 'red') + 'name:"%s" path:"%s" db:%d cache:%d>' % \
    (self.name, self.path, len(self), len(self.sqlite._cache[self.name]))

  def __repr__(self):
    return str(self)

  # ==================== Dictionary ==================== #
  def set(self, key, value):
    self.__setitem__(key, value)
    return self

  def __setitem__(self, key, value):
    with self.table_context():
      return self._sqlite.__setitem__(key, value)

  def __getitem__(self, key):
    with self.table_context():
      return self._sqlite.__getitem__(key)

  def update(self, items):
    with self.table_context():
      self._sqlite.update(items)
    return self

  def __iter__(self):
    with self.table_context():
      return self._sqlite.__iter__()

  def __contains__(self, key):
    with self.table_context():
      return self._sqlite.__contains__(key)

  def __len__(self):
    with self.table_context():
      return self._sqlite.__len__()

  def __delitem__(self, key):
    with self.table_context():
      return self._sqlite.__delitem__(key)

  def __cmp__(self, dict):
    if isinstance(dict, TableDict):
      return self._sqlite is dict._sqlite and \
          self._table_name == dict._table_name
    return False

  def keys(self):
    with self.table_context():
      return self._sqlite.keys()

  def values(self):
    with self.table_context():
      return self._sqlite.values()

  def items(self):
    with self.table_context():
      return self._sqlite.items()

  def clear(self):
    with self.table_context():
      self._sqlite.clear()
    return self


class SQLiteDict(NoSQL):
  """ Using SQLite in key-value pair manner

  Example
  -------
  >>> path = '/tmp/tmp.db'
  >>> db = SQLiteDict(path)
  >>> db1 = SQLiteDict(path)
  >>> print(id(db) == id(db1)) # True
  >>> def sett(i):
  ...     db = SQLiteDict(path)
  ...     db[i] = 'Process%d' % i
  ...     db.close()
  >>> def get(i):
  ...     db = SQLiteDict(path)
  ...     print("Get", db[i])
  >>> pros = [Process(target=sett, args=(i,))
  ...         for i in range(10)]
  >>> [p.start() for p in pros]
  >>> [p.join() for p in pros]
  ...
  >>> db = SQLiteDict(path)
  >>> print(db.items())
  >>> print(db[0])
  ...
  >>> pros = [Process(target=get, args=(i,))
  ...         for i in range(10)]
  >>> [p.start() for p in pros]
  >>> [p.join() for p in pros]

  Note
  ----
  numpy.ndarray will be converted to list before dumped to the database,
  hence, the speed of serialize and deserialize ndarray is extremely slow.
  This dict is purely performing in multiprocessing
  """

  _DEFAULT_TABLE = '_default_'

  def _restore_dict(self, path, read_only, cache_size):
    # specific cache dictionary for each table
    self._cache = defaultdict(dict)
    # ====== db manager ====== #
    self._conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES)
    self._conn.text_factory = str
    self._cursor = self._conn.cursor()
    # adjust pragma
    # SQLITE_OPEN_EXCLUSIVE
    self.connection.execute('PRAGMA main.locking_mode = EXCLUSIVE;')
    self.connection.execute("PRAGMA main.synchronous = 0;")
    self.connection.execute("PRAGMA journal_mode = MEMORY;")
    self.connection.commit()
    # ====== create default table ====== #
    self._current_table = SQLiteDict._DEFAULT_TABLE
    self.set_table(SQLiteDict._DEFAULT_TABLE)

  def _flush(self, save_all=False):
    curr_tab = self.current_table
    tables = self.get_all_tables() if save_all else [curr_tab]
    for tab in tables:
      self.set_table(tab)
      if len(self.current_cache) > 0:
        self.cursor.executemany(
            "INSERT INTO {tb} VALUES (?, ?)".format(tb=tab),
            [(str(k), _dump(v.tolist()) if isinstance(v, np.ndarray)
              else _dump(v))
             for k, v in self.current_cache.items()])
        self.connection.commit()
        self.current_cache.clear()
    # restore the last table
    return self.set_table(curr_tab)

  def _close(self):
    self._conn.close()

  # ==================== DB manager ==================== #
  def as_table(self, table_name):
    return TableDict(self, table_name)

  def set_table(self, table_name):
    if table_name is None:
      table_name = SQLiteDict._DEFAULT_TABLE
    table_name = str(table_name)
    if not self.is_table_exist(table_name):
      query = """CREATE TABLE {tb} (
                            key text NOT NULL,
                            value text NOT NULL,
                            PRIMARY KEY (key)
                        );"""
      self.cursor.execute(query.format(tb=table_name))
      self.connection.commit()
    # set the new table
    self._current_table = table_name
    return self

  @property
  def current_table(self):
    return self._current_table

  @property
  def current_cache(self):
    return self._cache[self.current_table]

  def get_all_tables(self):
    query = """SELECT name FROM sqlite_master where type='table';"""
    self.cursor.execute(query)
    return [table[0] for table in self.cursor.fetchall()]

  def drop_table(self, table_name=None):
    if self.read_only:
      return
    if table_name is None:
      table_name = self._current_table
      self._current_table = SQLiteDict._DEFAULT_TABLE
    else:
      table_name = str(table_name)
    if table_name == SQLiteDict._DEFAULT_TABLE:
      raise ValueError("Cannot drop default table.")
    self.cursor.execute("""DROP TABLE {tb};""".format(tb=table_name))
    return self

  def is_table_exist(self, table_name):
    try:
      self.cursor.execute('SELECT 1 FROM %s LIMIT 1;' % table_name)
      self.cursor.fetchone()
    except sqlite3.OperationalError as e:
      if "no such table" in str(e):
        return False
      else:
        raise e
    return True

  # ==================== pickling and properties ==================== #
  def __str__(self):
    curr_tab = self.current_table
    s = '<' + ctext('SQLiteDB:', 'red') + '%s>\n' % self.path
    all_tables = self.get_all_tables()
    name_fmt = '%' + str(max(len(i) for i in all_tables)) + 's'
    for tab in all_tables:
      self.set_table(tab)
      s += ' ' + ctext(name_fmt % tab, 'yellow') + ': db=%d cache=%d\n' % \
          (len(self), len(self.current_cache))
    self.set_table(curr_tab) # back to original table
    return s[:-1]

  @property
  def connection(self):
    return self._conn

  @property
  def cursor(self):
    return self._cursor

  # ==================== Dictionary ==================== #
  def __setitem__(self, key, value):
    if self.read_only:
      raise RuntimeError("Cannot __setitem__ for this Dict in read_only mode.")
    key = str(key)
    self.current_cache[key] = value
    if len(self.current_cache) >= self._cache_size:
      self.flush()
      self.current_cache.clear()

  def __getitem__(self, key):
    # ====== multiple keys select ====== #
    if isinstance(key, (tuple, list, np.ndarray)):
      query = """SELECT value FROM {tb}
                       WHERE key IN {keyval};"""
      keyval = '(' + ', '.join(['"%s"' % str(k) for k in key]) + ')'
      self.cursor.execute(
          query.format(tb=self._current_table, keyval=keyval))
      results = self.cursor.fetchall()
      # check if any not found keys
      if len(results) != len(key):
        raise KeyError("Cannot find all `key`='%s' in the dictionary." % keyval)
      # load binary data
      results = [marshal.loads(r[0]) for r in results]
    # ====== single key select ====== #
    else:
      key = str(key)
      if key in self.current_cache:
        return self.current_cache[key]
      query = """SELECT value FROM {tb} WHERE key="{keyval}" LIMIT 1;"""
      results = self.connection.execute(
          query.format(tb=self._current_table, keyval=key)).fetchone()
      # results = self.cursor.fetchone()
      if results is None:
        raise KeyError("Cannot find `key`='%s' in the dictionary." % key)
      results = marshal.loads(results[0])
    return results

  def __contains__(self, key):
    key = str(key)
    # check in cache
    if key in self.current_cache:
      return True
    # check in database
    query = """SELECT 1 FROM {tb} WHERE key="{keyval}" LIMIT 1;"""
    self.cursor.execute(
        query.format(tb=self._current_table, keyval=key))
    if self.cursor.fetchone() is None:
      return False
    return True

  def __len__(self):
    query = """SELECT COUNT(1) FROM {tb}""".format(tb=self._current_table)
    self.cursor.execute(query)
    n = self.cursor.fetchone()[0]
    return n + len(self.current_cache)

  def __delitem__(self, key):
    if self.read_only:
      return
    query = """DELETE FROM {tb} WHERE {cond};"""
    if isinstance(key, (tuple, list, Iterator, np.ndarray)):
      key = [str(k) for k in key]
    else:
      key = [str(key)]
    # ====== check if key in cache ====== #
    db_key = []
    for k in key:
      if k in self.current_cache:
        del self.current_cache[k]
      else:
        db_key.append(k)
    # ====== remove key from db ====== #
    self.cursor.execute(
        query.format(tb=self._current_table,
                     cond='key IN ("%s")' % ', '.join(db_key)))
    self.connection.commit()

  def keys(self):
    for k in self.cursor.execute(
        """SELECT key from {tb};""".format(tb=self._current_table)):
      yield k[0]
    for k in self.current_cache.keys():
      yield k

  def values(self):
    for val in self.cursor.execute(
        """SELECT value from {tb};""".format(tb=self._current_table)):
      yield marshal.loads(val[0])
    for v in self.current_cache.values():
      yield v

  def items(self):
    for item in self.cursor.execute(
        """SELECT key, value from {tb};""".format(tb=self._current_table)):
      yield (item[0], marshal.loads(item[1]))
    for k, v in self.current_cache.items():
      yield k, v

  def update(self, items):
    if self.read_only:
      return
    query = """UPDATE {tb} SET value=(?) WHERE key=("?");"""
    if isinstance(items, Mapping):
      items = items.items()
    # ====== check if update is in cache ====== #
    db_update = []
    for key, value in items:
      key = str(key)
      if key in self.current_cache:
        self.current_cache[key] = value
      else:
        db_update.append((_dump(value), key))
    # ====== perform DB update ====== #
    self.cursor.executemany(query.format(tb=self._current_table), db_update)
    self.connection.commit()
    return self

  def clear(self):
    if self.read_only:
      return
    self.cursor.execute("""TRUNCATE TABLE {tb};""".format(tb=self._current_table))
    self.connection.commit()
    self.current_cache.clear()
    return self



# -*- coding: utf-8 -*-
"""
Note
----
The following order is recommended for extracting spectra:
+ AudioReader:
    - Loading raw audio
    - remove DC offeset and dithering
    - preemphasis
+ SpectraExtractor (or CQTExtractor):
    - Extracting the Spectra
# ===== SAD ===== #
+ SADgmm:
    - Extracting SAD (optional)
+ Read3ColSAD:
    - Generating SAD labels from 3-cols files
# ===== Normalization ===== #
+ Rastafilt:
    - Rastafilt (optional for MFCC)
+ DeltaExtractor (calculated before applying SAD)
    - Calculate Deltas (and shifted delta for MFCCs).
+ ApplyingSAD:
    - Applying SAD indexing on features
+ BNFExtractor:
    - Extracting bottleneck features
+ AcousticNorm
    - Applying CMVN and WCMVN (This is important so the SAD frames
    are not affected by the nosie).
"""
# ===========================================================================
# The waveform and spectrogram preprocess utilities is adapted from:
# [librosa](https://github.com/bmcfee/librosa)
# Copyright (c) 2016, librosa development team.
# Modified work Copyright 2016-2017 TrungNT
# ===========================================================================
from __future__ import print_function, division, absolute_import

import os
import re
import six
import math
import copy
import base64
import shutil
import inspect
import warnings
from numbers import Number
from six import string_types
from collections import OrderedDict, Mapping, defaultdict

import numpy as np
import tensorflow as tf
from scipy.signal import lfilter

from odin.fuel import Dataset, MmapData, MmapDict
from odin.utils import (is_number, cache_memory, is_string, as_tuple,
                        get_all_files, is_pickleable, Progbar, mpi, ctext,
                        is_fileobj, batching)
from odin.preprocessing.base import Extractor, ExtractorSignal
from odin.preprocessing.signal import (smooth, pre_emphasis, get_window, get_energy,
                                       spectra, vad_energy,
                                       pitch_track, resample, rastafilt, mvn, wmvn,
                                       shifted_deltas, stack_frames, stft,
                                       power_spectrogram, mels_spectrogram, ceps_spectrogram,
                                       power2db, anything2wav)
# import all OpenSMILE extractor
from odin.preprocessing._opensmile import *

# ===========================================================================
# Predefined variables of speech datasets
# ===========================================================================
# ==================== Timit ==================== #
timit_61 = ['aa', 'ae', 'ah', 'ao', 'aw', 'ax', 'ax-h', 'axr', 'ay',
    'b', 'bcl', 'ch', 'd', 'dcl', 'dh', 'dx', 'eh', 'el', 'em', 'en',
    'eng', 'epi', 'er', 'ey', 'f', 'g', 'gcl', 'h#', 'hh', 'hv', 'ih',
    'ix', 'iy', 'jh', 'k', 'kcl', 'l', 'm', 'n', 'ng', 'nx', 'ow',
    'oy', 'p', 'pau', 'pcl', 'q', 'r', 's', 'sh', 't', 'tcl', 'th',
    'uh', 'uw', 'ux', 'v', 'w', 'y', 'z', 'zh']
timit_39 = ['aa', 'ae', 'ah', 'aw', 'ay', 'b', 'ch', 'd',
    'dh', 'dx', 'eh', 'er', 'ey', 'f', 'g', 'hh', 'ih', 'iy', 'jh', 'k',
    'l', 'm', 'n', 'ng', 'ow', 'oy', 'p', 'r', 's', 'sh', 'sil', 't',
    'th', 'uh', 'uw', 'v', 'w', 'y', 'z']
timit_map = {'ao': 'aa', 'ax': 'ah', 'ax-h': 'ah', 'axr': 'er',
    'hv': 'hh', 'ix': 'ih', 'el': 'l', 'em': 'm',
    'en': 'n', 'nx': 'n',
    'eng': 'ng', 'zh': 'sh', 'ux': 'uw',
    'pcl': 'sil', 'tcl': 'sil', 'kcl': 'sil', 'bcl': 'sil',
    'dcl': 'sil', 'gcl': 'sil', 'h#': 'sil', 'pau': 'sil', 'epi': 'sil'}

# ===========================================================================
# Basics
# ===========================================================================
def _read_pcm(path, encode):
  dtype = np.int16
  sr = None
  if encode is not None:
    if 'ulaw' in encode.lower():
      dtype = np.int8
      sr = 8000
    elif 'vast' in encode.lower():
      dtype = np.int16
      sr = 44000
  raw = np.memmap(path, dtype=dtype, mode='r')
  return raw, sr

def read(path_or_file, encode=None):
  """
  Returns
  -------
  audio_array : [n_samples, nb_channels]
    the audio array
  sr : {int, None}
    sample rate
  """
  # ====== check input ====== #
  if is_fileobj(path_or_file):
    f = path_or_file
    f.seek(0)
    path = f.name
  elif os.path.isfile(path_or_file):
    f = None
    path = path_or_file
  else:
    raise ValueError("Invalid type of `path_or_file` %s" %
        str(type(path_or_file)))
  # ====== read the audio ====== #
  if '.pcm' in path.lower():
    f = open(path_or_file, 'rb')
    raw, sr = _read_pcm(f, encode=encode)
  else:
    import soundfile
    try:
      f = open(path_or_file, 'rb')
      raw, sr = soundfile.read(f)
    except Exception as e:
      # read special pcm file
      if '.sph' in f.name.lower():
        f = open(path_or_file, 'rb')
        raw, sr = _read_pcm(f, encode=encode)
      # read using external tools
      else:
        raw, sr = anything2wav(inpath=path, outpath=None,
                               codec=encode, return_data=True)
  # close file
  if f is not None:
    f.close()
  return raw, sr

def save(file_or_path, s, sr, subtype=None):
  '''
  Parameters
  ----------
  s : array_like
      The data to write.  Usually two-dimensional (channels x frames),
      but one-dimensional `data` can be used for mono files.
      Only the data types ``'float64'``, ``'float32'``, ``'int32'``
      and ``'int16'`` are supported.

      .. note:: The data type of `data` does **not** select the data
                type of the written file. Audio data will be
                converted to the given `subtype`. Writing int values
                to a float file will *not* scale the values to
                [-1.0, 1.0). If you write the value ``np.array([42],
                dtype='int32')``, to a ``subtype='FLOAT'`` file, the
                file will then contain ``np.array([42.],
                dtype='float32')``.
  subtype: str
      'PCM_24': 'Signed 24 bit PCM'
      'PCM_16': 'Signed 16 bit PCM'
      'PCM_S8': 'Signed 8 bit PCM'

  Return
  ------
  waveform (ndarray), sample rate (int)
  '''
  from soundfile import write
  return write(file_or_path, s, sr, subtype=subtype)

# ===========================================================================
# Helper function
# ===========================================================================
def _extract_frame_step_length(sr, frame_length, step_length):
  # ====== check frame length ====== #
  if frame_length < 1.:
    frame_length = int(sr * frame_length)
  else:
    frame_length = int(frame_length)
  # ====== check step length ====== #
  if step_length is None:
    step_length = frame_length // 4
  elif step_length < 1.:
    step_length = int(sr * step_length)
  else:
    step_length = int(step_length)
  return frame_length, step_length

@cache_memory
def _num_two_factors(x):
  """return number of times x is divideable for 2"""
  if x <= 0:
    return 0
  num_twos = 0
  while x % 2 == 0:
    num_twos += 1
    x //= 2
  return num_twos

@cache_memory
def _max_fft_bins(sr, n_fft, fmax):
  return [i + 1 for i, j in enumerate(np.linspace(0, float(sr) / 2, int(1 + n_fft // 2),
                                      endpoint=True)) if j >= fmax][0]

def audio_segmenter(files, outpath, max_duration,
                    sr=None, sr_new=None, best_resample=True,
                    override=False):
  """ Segment all given files into small chunks, the new file
  name is formatted as:
   - [name_without_extension].[ID].wav

  The information for each segment is saved at a csv file:
   - [outpath]/segments.csv

  Note
  ----
  We separated the segmenter from FeatureProcessor, since you can try
  different configuration for the features on the same set of segments,
  it is efficient to do this once-for-all.
  """
  info_path = os.path.join(outpath, 'segments.csv')
  # ====== validate arguments ====== #
  max_duration = int(max_duration)
  files = [f for f in as_tuple(files, t=str)
           if os.path.isfile(f)]
  outpath = str(outpath)
  if os.path.isfile(outpath):
    raise ValueError("outpath at: %s is a file." % outpath)
  if os.path.isdir(outpath):
    if not override:
      return info_path
    else:
      shutil.rmtree(outpath)
  if not os.path.isdir(outpath):
    os.mkdir(outpath)
  reader = AudioReader(sr=sr, sr_new=sr_new, best_resample=best_resample,
                       remove_dc_n_dither=False, preemphasis=None)

  # ====== segmenting ====== #
  def segmenting(f):
    raw = reader.transform(f)
    path, sr, raw = raw['path'], raw['sr'], raw['raw']
    segs = [int(np.round(i)) for i in np.linspace(
        start=0, stop=raw.shape[0],
        num=int(np.ceil(raw.shape[0] / (sr * max_duration))) + 1,
        endpoint=True)]
    indices = list(zip(segs, segs[1:]))
    name = os.path.basename(path)
    info = []
    for idx, (s, e) in enumerate(indices):
      y = raw[s:e]
      seg_name = name.split('.')[:-1] + [str(idx), 'wav']
      seg_name = '.'.join(seg_name)
      save(os.path.join(outpath, seg_name), y, sr)
      info.append((seg_name, s / sr, e / sr))
    return path, info
  nb_files = len(files)
  prog = Progbar(target=nb_files, print_summary=True, print_report=True,
                 name='Segmenting to path: %s' % outpath)
  task = mpi.MPI(jobs=files, func=segmenting,
                 ncpu=None, batch=1, backend='python')
  # ====== running the processor ====== #
  seg_indices = []
  for f, info in task:
    prog['File'] = f
    prog['#Segs'] = len(info)
    assert all(e - s <= max_duration
               for name, s, e in info), \
        "Results contain segments > max duration, file: %s, segs: %s" %\
        (f, str(info))
    for seg, s, e in info:
      seg_indices.append((seg, os.path.basename(f), s, e))
    prog.add(1)
  # ====== save the info ====== #
  header = ' '.join(['segment', 'origin', 'start', 'end'])
  np.savetxt(info_path, seg_indices,
             fmt='%s', delimiter=' ', header=header, comments='')
  print("Segment info saved at:", ctext(info_path, 'cyan'))
  return info_path

# ===========================================================================
# I/O
# ===========================================================================
class AudioReader(Extractor):

  """ Return a dictionary of
  {
      'raw': loaded_signal,
      'duration': in second,
      'sr': sample rate,
      'path': path_to_loaded_file
  }

  Parameters
  ----------
  sr: int or None
      provided sr for missing sr audio (i.e. pcm files),
      NOTE this value only used when cannot find `sr` information
      from audio file (example: reading raw pcm).
  sr_new: int or None
      resample sample rate for all provided audio, only
      support downsample (i.e. must be smaller than sr).
  remove_dc_n_dither : bool
    dithering adds noise to the signal to remove periodic noise

  Input
  -----
  path_or_array: string, tuple, list, mapping
      - string for path
      - tuple or list for (path-or-raw, sr)
      - mapping for provding additional information include:
      sr, encode (ulaw, vast), 'raw' or 'path'

  Note
  ----
  Dithering introduces white noise when you save the raw array into
  audio file.
  For now only support one channel
  """

  def __init__(self, sr=None, sr_new=None, best_resample=True,
               remove_dc=True, dataset=None):
    super(AudioReader, self).__init__(is_input_layer=True)
    self.sr = sr
    self.sr_new = sr_new
    self.best_resample = best_resample
    self.remove_dc = bool(remove_dc)
    # ====== check dataset ====== #
    if is_string(dataset) and os.path.isdir(dataset):
      dataset = Dataset(dataset)
    elif dataset is not None and not isinstance(dataset, Dataset):
      raise ValueError("dataset can be instance of odin.fuel.Dataset or None")
    self.dataset = dataset

  def _transform(self, path_or_array):
    raw = None # raw
    sr = None # by default, we don't know sample rate
    name = None
    path = None
    duration = None
    encode = None
    channel = None
    # ====== check path_or_array ====== #
    # mapping of specific data
    if isinstance(path_or_array, Mapping):
      if 'sr' in path_or_array:
        sr = int(path_or_array['sr'])
      if 'encode' in path_or_array:
        encode = str(path_or_array['encode'])
      if 'channel' in path_or_array:
        channel = int(path_or_array['channel'])
      # get raw or path out of the Dictionary
      if 'raw' in path_or_array:
        raw = path_or_array['raw']
      elif 'path' in path_or_array:
        path = str(path_or_array['path'])
        raw, sr = read(path, encode=encode)
      else:
        raise ValueError('`path_or_array` can be a dictionary, contains '
            'following key: sr, raw, path. One of the key `raw` for '
            'raw array signal, or `path` for path to audio file must '
            'be specified.')
    # read string file pth
    elif is_string(path_or_array):
      path = path_or_array
      if os.path.isfile(path_or_array):
        raw, sr = read(path_or_array, encode=encode)
      # given a dataset
      elif self.dataset is not None:
        start, end = self.dataset['indices'][path_or_array]
        raw = self.dataset['raw'][start:end]
        sr = int(self.dataset['sr'][path_or_array])
        name = path_or_array
        if 'path' in self.dataset:
          path = self.dataset['path'][path_or_array]
      # file not exist
      else:
        raise ValueError("Cannot locate file at path: %s" % path_or_array)
    # read from file object
    elif is_fileobj(path_or_array):
      path = path_or_array.name
      raw, sr = read(path_or_array, encode=encode)
    else:
      raise ValueError("`path_or_array` can be: list, tuple, Mapping, string, file"
          ". But given: %s" % str(type(path_or_array)))
    # ====== check channel ====== #
    raw = raw.astype('float32')
    if raw.ndim == 1:
      pass
    elif raw.ndim == 2:
      if raw.shape[0] == 2:
        raw = raw.T
      if channel is not None:
        raw = raw[:, channel]
    else:
      raise ValueError("No support for %d-D signal from file: %s" %
          (raw.ndim, path))
    # ====== valiate sample rate ====== #
    if sr is None and self.sr is not None:
      sr = int(self.sr)
    # resampling if necessary
    if sr is not None and self.sr_new is not None:
      raw = resample(raw, sr, self.sr_new,
                     best_algorithm=self.best_resample)
      sr = int(self.sr_new)
    # ====== remove DC offset ====== #
    if self.remove_dc:
      raw = raw - np.mean(raw, 0).astype(raw.dtype)
    # ====== get duration if possible ====== #
    if sr is not None:
      duration = max(raw.shape) / sr
    # ====== check if absolute path====== #
    if path is not None:
      if not is_string(path):
        path = str(path, 'utf-8')
      if '/' != path[0]:
        path = os.path.abspath(path)
    ret = {'raw': raw, 'sr': sr, 'duration': duration, # in second
           'path': path}
    if name is not None:
      ret['name'] = name
    return ret

class AudioAugmentor(Extractor):
  """ SREAugmentor

  New name for each utterance is:
    [utt_name]/[noise1_name]/[noise2_name]...
  """

  def __init__(self, musan_path, rirs_path):
    super(AudioAugmentor, self).__init__(is_input_layer=False)
    # TODO: finish this

  def _transform(self, row):
    pass

# ===========================================================================
# Pre-processing raw signal
# ===========================================================================
class Dithering(Extractor):
  """ Dithering """

  def __init__(self, input_name=('raw', 'sr'), output_name='raw'):
    super(Dithering, self).__init__(
        input_name=as_tuple(input_name, t=string_types),
        output_name=str(output_name))

  def _transform(self, feat):
    raw, sr = [feat[name] for name in self.input_name]
    # assuming 16-bit
    if max(abs(raw)) <= 1.:
      raw = raw * 2**15
    # select alpha
    if sr == 16000:
      alpha = 0.99
    elif sr == 8000:
      alpha = 0.999
    else:
      raise ValueError('Sampling frequency %s not supported' % sr)
    slen = raw.size
    raw = lfilter([1, -1], [1, -alpha], raw)
    dither = np.random.rand(slen) + np.random.rand(slen) - 1
    s_pow = max(raw.std(), 1e-20)
    raw = raw + 1.e-6 * s_pow * dither
    return {self.output_name: raw}

class PreEmphasis(Extractor):
  """ PreEmphasis

  Parameters
  ----------
  coeff : float (0-1)
      pre-emphasis filter, if 0 or None, no filter applied
  input_name : string
      name of raw signal in the features pipeline dictionary

  """

  def __init__(self, coeff=0.97,
               input_name='raw', output_name='raw'):
    super(PreEmphasis, self).__init__(input_name=str(input_name),
                                      output_name=str(output_name))
    assert 0. < coeff < 1.
    self.coeff = float(coeff)

  def _transform(self, feat):
    raw = feat[self.input_name]
    if not 0 < raw.ndim <= 2:
      raise ValueError("Only supper 1 or 2 channel audio but given shape: %s" %
                       str(raw.shape))
    return {self.output_name: pre_emphasis(raw, coeff=self.coeff)}

# ===========================================================================
# Low-level operator
# ===========================================================================
class Framing(Extractor):
  """ Framing

  Parameters
  ----------

  """

  def __init__(self, frame_length, step_length=None,
               window='hamm', padding=False,
               input_name=('raw', 'sr'), output_name='frames'):
    if isinstance(input_name, string_types):
      input_name = (input_name, 'sr')
    assert isinstance(output_name, string_types), "`output_name` must be string"
    super(Framing, self).__init__(input_name=input_name, output_name=output_name)
    if step_length is None:
      step_length = frame_length // 4
    self.frame_length = frame_length
    self.step_length = step_length
    self.window = window
    self.padding = bool(padding)

  def _transform(self, y_sr):
    y, sr = [y_sr[name] for name in self.input_name]
    frame_length, step_length = _extract_frame_step_length(
        sr, self.frame_length, self.step_length)
    # ====== check if padding zeros ====== #
    if self.padding:
      y = np.pad(y, int(frame_length // 2), mode='constant')
    # ====== framing the signal ====== #
    shape = y.shape[:-1] + (y.shape[-1] - frame_length + 1, frame_length)
    strides = y.strides + (y.strides[-1],)
    y_frames = np.lib.stride_tricks.as_strided(y, shape=shape, strides=strides)
    if y_frames.ndim > 2:
      y_frames = np.rollaxis(y_frames, 1)
    y_frames = y_frames[::step_length] # [n, frame_length]
    # ====== prepare the window function ====== #
    if self.window is not None:
      fft_window = get_window(
          self.window, frame_length, periodic=True).reshape(1, -1)
      y_frames = fft_window * y_frames
      # scaling the windows
      scale = np.sqrt(1.0 / fft_window.sum()**2)
    else:
      scale = np.sqrt(1.0 / frame_length**2)
    # ====== calculate frames energy ====== #
    return {self.output_name: y_frames,
            'scale': scale}

class CalculateEnergy(Extractor):
  """
  Parameters
  ----------
  log : bool (default: True)
    take the natural logarithm of the energy

  Input
  -----
  numpy.ndarray : 2D [n_samples, n_features]

  Output
  ------
  numpy.ndarray : 1D [n_samples,]
    Energy for each frames

  """

  def __init__(self, log=True,
               input_name='frames', output_name='energy'):
    super(CalculateEnergy, self).__init__(input_name=str(input_name),
                                  output_name=str(output_name))
    self.log = bool(log)

  def _transform(self, X):
    frames = X[self.input_name]
    energy = get_energy(frames, log=self.log).astype('float32')
    return {self.output_name: energy}

# ===========================================================================
# Spectrogram
# ===========================================================================
class STFTExtractor(Extractor):
  """ Short time Fourier transform
  `window` should be `None` if input is windowed framed signal

  Parameters
  ----------
  frame_length: {int, float}
      number of samples point for 1 frame, or length of frame in millisecond
  step_length: {int, float}
      number of samples point for 1 step (when shifting the frames),
      or length of step in millisecond
      If unspecified, defaults `win_length / 4`.
  n_fft: int > 0 [scalar]
      FFT window size
      If not provided, uses the smallest power of 2 enclosing `frame_length`.

  scale : {None, string, float}
      value for scaling the matrix after STFT, important for
      iSTFT afterward
      if None, no extra scale is performed
      if string, looking for feature with given name in the pipeline
      if float, directly using given value for scaling

  Input
  -----
  numpy.ndarray : [n_samples,] or [n_frames, frame_length]
    raw signal or framed signal
  integer : > 0
    sample rate of the audio

  Output
  ------
  'stft' : complex64 array [time, frequency]
  'stft_energy' : float32 array [time, 1]
  """

  def __init__(self, frame_length=None, step_length=None, n_fft=512,
               window='hamm', padding=False, energy=True, scale=None,
               input_name=('raw', 'sr'), output_name='stft'):
    if isinstance(input_name, string_types):
      input_name = (input_name, 'sr')
    assert isinstance(output_name, string_types), "`output_name` must be string"
    super(STFTExtractor, self).__init__(input_name=input_name,
                                        output_name=output_name)
    self.frame_length = frame_length
    self.step_length = step_length
    self.n_fft = n_fft
    self.window = window
    self.padding = bool(padding)
    self.energy = bool(energy)
    assert isinstance(scale, (string_types, Number, type(None)))
    self.scale = scale

  def _transform(self, y_sr):
    y, sr = [y_sr[name] for name in self.input_name]
    scale = self.scale
    if isinstance(scale, string_types):
      scale = y_sr[scale]
    # ====== check frame_length ====== #
    if self.frame_length is None:
      if y.ndim == 2 and y.shape[1] > 2:
        frame_length = y.shape[1]
        step_length = None
      else:
        raise ValueError("`frame_length` is not provided, the input to "
                         "the extractor must be framed signal from "
                         "`odin.preprocessing.speech.Framing`")
    else:
      frame_length, step_length = _extract_frame_step_length(
          sr, self.frame_length, self.step_length)
    # ====== stft ====== #
    results = stft(y=y,
                   frame_length=frame_length, step_length=step_length,
                   n_fft=self.n_fft, window=self.window, scale=scale,
                   padding=self.padding, energy=self.energy)
    if self.energy:
      s, e = results
      return {self.output_name: s,
              '%s_energy' % self.output_name: e}
    else:
      return {self.output_name: results}

class PowerSpecExtractor(Extractor):
  """ Extract power spectrogram from complex STFT array

  Output
  ------
  'spec' : [time, n_fft / 2 + 1]

  """

  def __init__(self, power=2.0,
               input_name='stft', output_name='spec'):
    super(PowerSpecExtractor, self).__init__(input_name=input_name,
                                             output_name=output_name)
    self.power = float(power)

  def _transform(self, X):
    return power_spectrogram(S=X[self.input_name], power=self.power)

class MelsSpecExtractor(Extractor):
  """
  Parameters
  ----------
  input_name : (string, string) (default: ('spec', 'sr'))
    the name of spectrogram and sample rate in the feature pipeline

  Output
  ------
  'mspec' : [time, n_mels]

  """

  def __init__(self, n_mels, fmin=64, fmax=None, top_db=80.0,
               input_name=('spec', 'sr'), output_name='mspec'):
    # automatically add sample rate to input_name
    if isinstance(input_name, string_types):
      input_name = (input_name, 'sr')
    super(MelsSpecExtractor, self).__init__(input_name=input_name,
                                            output_name=output_name)
    self.n_mels = int(n_mels)
    self.fmin = fmin
    self.fmax = fmax
    self.top_db = top_db

  def _transform(self, X):
    return mels_spectrogram(spec=X[self.input_name[0]], sr=X[self.input_name[1]],
                            n_mels=self.n_mels,
                            fmin=self.fmin, fmax=self.fmax, top_db=self.top_db)

class MFCCsExtractor(Extractor):
  """
  """

  def __init__(self, n_ceps,
               remove_first_coef=True, first_coef_energy=False,
               input_name='mspec', output_name='mfcc'):
    super(MFCCsExtractor, self).__init__(input_name=input_name,
                                         output_name=output_name)
    self.n_ceps = int(n_ceps)
    self.remove_first_coef = bool(remove_first_coef)
    self.first_coef_energy = bool(first_coef_energy)

  def _transform(self, X):
    n_ceps = self.n_ceps
    if self.remove_first_coef:
      n_ceps += 1
    mfcc = ceps_spectrogram(mspec=X[self.input_name],
                            n_ceps=n_ceps,
                            remove_first_coef=False)
    ret = {
        self.output_name: mfcc[:, 1:] if self.remove_first_coef else mfcc}
    if self.first_coef_energy:
      ret['%s_energy' % self.output_name] = mfcc[:, 0]
    return ret

class Power2Db(Extractor):
  """ Convert power spectrogram to Decibel spectrogram

  """

  def __init__(self, input_name, output_name=None, top_db=80.0):
    input_name = as_tuple(input_name, t=string_types)
    super(Power2Db, self).__init__(input_name=input_name,
                                   output_name=output_name)
    self.top_db = float(top_db)

  def _transform(self, X):
    return [power2db(S=X[name], top_db=self.top_db) for name in self.input_name]

class SpectraExtractor(Extractor):
  """AcousticExtractor

  Parameters
  ----------
  frame_length: {int, float}
      number of samples point for 1 frame, or length of frame in millisecond
  step_length: {int, float}
      number of samples point for 1 step (when shifting the frames),
      or length of step in millisecond
      If unspecified, defaults `win_length / 4`.
  n_fft: int > 0 [scalar]
      FFT window size
      If not provided, uses the smallest power of 2 enclosing `frame_length`.
  window : string, tuple, number, function, or np.ndarray [shape=(n_fft,)]
      - a window specification (string, tuple, or number);
        see `scipy.signal.get_window`
      - a window function, such as `scipy.signal.hanning`
      - a vector or array of length `n_fft`
  power : float > 0 [scalar]
      Exponent for the magnitude spectrogram.
      e.g., 1 for energy (or magnitude), 2 for power, etc.
  log: bool
      if True, convert all power spectrogram to DB
  padding : bool
      - If `True`, the signal `y` is padded so that frame
        `D[:, t]` is centered at `y[t * hop_length]`.
      - If `False`, then `D[:, t]` begins at `y[t * hop_length]`
  """

  def __init__(self, frame_length, step_length=None, n_fft=512, window='hann',
               n_mels=None, n_ceps=None, fmin=64, fmax=None,
               power=2.0, log=True, padding=False,
               input_name=('raw', 'sr')):
    super(SpectraExtractor, self).__init__(input_name=input_name)
    # ====== STFT ====== #
    self.frame_length = frame_length
    self.step_length = step_length
    self.n_fft = n_fft
    self.window = window
    # ====== ceptral analysis ====== #
    self.n_mels = n_mels
    self.n_ceps = n_ceps
    self.fmin = fmin
    self.fmax = fmax
    # ====== power spectrum ====== #
    self.power = float(power)
    self.log = bool(log)
    # ====== others ====== #
    self.padding = bool(padding)

  def _transform(self, y_sr):
    y, sr = [y_sr[i] for i in self.input_name]
    frame_length, step_length = _extract_frame_step_length(
        sr, self.frame_length, self.step_length)
    # ====== extract spectra ====== #
    feat = spectra(sr=sr, frame_length=frame_length, y=y, S=None,
                   step_length=step_length, n_fft=self.n_fft,
                   window=self.window,
                   n_mels=self.n_mels, n_ceps=self.n_ceps,
                   fmin=self.fmin, fmax=self.fmax,
                   top_db=80., power=self.power, log=self.log,
                   padding=self.padding)
    return feat

class CQTExtractor(Extractor):
  """ Constant-Q transform
  Using log-scale instead of linear-scale frequencies for
  signal analysis

  """

  def __init__(self, frame_length, step_length=None, n_bins=96, window='hann',
               n_mels=None, n_ceps=None, fmin=64, fmax=None, padding=False,
               input_name=('raw', 'sr')):
    super(CQTExtractor, self).__init__(input_name=input_name)
    self.frame_length = frame_length
    self.step_length = step_length
    self.n_bins = int(n_bins)
    self.window = window
    self.n_mels = n_mels
    self.n_ceps = n_ceps
    self.fmin = fmin
    self.fmax = fmax
    self.padding = padding

  def _transform(self, y_sr):
    y, sr = [y_sr[name] for name in self.input_name]
    frame_length, step_length = _extract_frame_step_length(
        sr, self.frame_length, self.step_length)
    # ====== extract CQT ====== #
    from librosa.core import constantq
    # auto adjust bins_per_octave to get maximum range of frequency
    bins_per_octave = np.ceil(float(self.n_bins - 1) / np.log2(sr / 2. / self.fmin)) + 1
    # adjust the bins_per_octave to make acceptable hop_length
    # i.e. 2_factors(hop_length) < [ceil(cqt_bins / bins_per_octave) - 1]
    if _num_two_factors(step_length) < np.ceil(self.n_bins / bins_per_octave) - 1:
      bins_per_octave = np.ceil(self.n_bins / (_num_two_factors(step_length) + 1))
    with warnings.catch_warnings():
      warnings.filterwarnings("ignore", category=DeprecationWarning)
      qtrans = constantq.cqt(y=y, sr=sr, hop_length=step_length, n_bins=self.n_bins,
                             bins_per_octave=int(bins_per_octave),
                             fmin=self.fmin, tuning=0.0, norm=1,
                             filter_scale=1., sparsity=0.01).astype('complex64')
    # ====== ceptral analysis ====== #
    feat = spectra(sr=sr, frame_length=frame_length, y=None, S=qtrans.T,
                   step_length=step_length, n_fft=None, window='hann',
                   n_mels=self.n_mels, n_ceps=self.n_ceps,
                   fmin=64, fmax=self.fmax,
                   top_db=80.0, power=2.0, log=True,
                   padding=self.padding)
    # ====== add 'q' prefix for CQT features ====== #
    feat = {'q' + name: X for name, X in feat.items()}
    return feat

# ===========================================================================
# Bottleneck features
# ===========================================================================
class _BNFExtractorBase(Extractor):
  """ _BNFExtractorBase """

  def __init__(self, input_name, network, output_name='bnf',
               sad_name='sad', remove_non_speech=True,
               stack_context=10, pre_mvn=True,
               batch_size=2048):
    assert isinstance(input_name, string_types), "`input_name` must be string"
    if isinstance(sad_name, string_types):
      input_name = (input_name, sad_name)
      self.use_sad = True
    else:
      self.use_sad = False
    self.remove_non_speech = bool(remove_non_speech)
    super(_BNFExtractorBase, self).__init__(
        input_name=input_name, output_name=output_name)
    # ====== other configs ====== #
    if stack_context is None:
      stack_context = 0
    self.stack_context = int(stack_context)
    self.pre_mvn = bool(pre_mvn)
    self.batch_size = int(batch_size)
    # ====== check the network ====== #
    self._prepare_network(network)

  def _prepare_input(self, X, sad):
    # ====== pre-normalization ====== #
    X_sad = X[sad] if sad is not None else X
    if self.pre_mvn:
      X = (X - X_sad.mean(0, keepdims=True)) /  \
          (X_sad.std(0, keepdims=True) + 1e-18)
    # ====== stacking context and applying SAD ====== #
    if self.stack_context > 0:
      X = stack_frames(X, frame_length=self.stack_context * 2 + 1,
                       step_length=1, keep_length=True,
                       make_contigous=True)
    if self.remove_non_speech and sad is not None:
      X = X[sad]
    return X

  def _transform(self, feat):
    if self.use_sad:
      X, sad = feat[self.input_name[0]], feat[self.input_name[1]]
      sad = sad.astype('bool')
      assert len(sad) == len(X), \
      "Input mismatch, `sad` has shape: %s, and input feature with name: %s; shape: %s" % \
      (sad.shape, self.input_name[0], X.shape)
    else:
      X = feat[self.input_name]
      sad = None
    # ====== transform ====== #
    X = self._prepare_input(X, sad)
    y = []
    # make prediction
    for s, e in batching(n=X.shape[0], batch_size=self.batch_size):
      y.append(self._apply_dnn(X[s:e]))
    return np.concatenate(y, axis=0)

  def _prepare_network(self, network):
    raise NotImplementedError

  def _apply_dnn(self, X):
    raise NotImplementedError

class BNFExtractorCPU(_BNFExtractorBase):
  """ Deep bottleneck feature extractor
  The following order of preprocessing features for BNF are suggested:
  * extract input features: `X`
  * extract sad: `S`
  * normalize by speech activities: `X = (X - X[S].mean()) / X[S].std()`
  * Stacking features: `X = stacking(X, context=10)`
  => X_bnf = network(X)

  Parameters
  ----------
  input_feat : str
    name of input feature
  network : odin.nnet.base.NNOp
    instance pre-trained NNOp
  pre_mvn : bool (default: False)
    perform mean-variance normalization before stacking,
    then, feeding data to network.
  sad_name : {str, None}
    if None, or `sad_name` not found, don't applying SAD to
    the input feature before BNF
  remove_non_speech : bool (default: True)
    if True, remove non-speech frames using given SAD
  batch_size : int
    batch size when feeding data to the network, suggest
    to have as much data as possible.

  Note
  ----
  Order of preprocessing for BNF:
   - delta and delta delta extraction (optional)
   - mean_var_norm based on SAD frames statistics.
   - Stacking the left and right context frames.
   - Applying SAD indices.
   - Mean-variance normalization
   => BNFExtractor

  """

  def _prepare_network(self, network):
    from odin.nnet.models.bnf import _BNFbase
    if isinstance(network, _BNFbase):
      network = network.__class__
    assert isinstance(network, type) and issubclass(network, _BNFbase), \
    "`network` must be a subclass of odin.nnet.models.Model, but given: %s"\
    % str(network)
    params = network.load_parameters()
    # note: the weights are transposed for column matrix (Matlab format)
    self.weights = [params[name][:]
        for name in sorted([key for key in params.keys() if 'w' == key[0]])]
    self.biases = [params[name][:]
        for name in sorted([key for key in params.keys() if 'b' == key[0]])]
    assert len(self.weights) == len(self.biases), \
    'Number of weights is: %d; but number of biases is: %s' % \
    (len(self.weights), len(self.biases))

  def _renorm_rms(self, x, target_rms=1.0, axis=0):
    """ scales the data such that RMS is 1.0 """
    # scale = sqrt(x^t x / (D * target_rms^2)).
    D = np.sqrt(x.shape[axis])
    x_rms = np.sqrt(np.sum(x * x, axis=axis, keepdims=True)) / D
    x_rms[x_rms == 0] = 1.
    return target_rms * x / x_rms

  def _apply_dnn(self, X):
    assert X.shape[1] == self.weights[0].shape[1], \
    "Input must has dimension (?, %d) but given tensor with shape: %s" % \
    (self.weights[0].shape[0], str(X.shape))
    X = X.T
    for wi, bi in zip(self.weights[:-1], self.biases[:-1]):
      X = wi.dot(X) + bi
      # relu nonlinearity
      np.maximum(X, 0, out=X)
      # scale the RMS
      X = self._renorm_rms(X, axis=0)
    # last layer, only linear
    X = self.weights[-1].dot(X) + self.biases[-1]
    return X.T

class BNFExtractor(_BNFExtractorBase):
  """ Deep bottleneck feature extractor
  The following order of preprocessing features for BNF are suggested:
  * extract input features: `X`
  * extract sad: `S`
  * normalize by speech activities: `X = (X - X[S].mean()) / X[S].std()`
  * Stacking features: `X = stacking(X, context=10)`
  => X_bnf = network(X)

  Parameters
  ----------
  input_feat : str
    name of input feature
  network : odin.nnet.base.NNOp
    instance pre-trained NNOp
  pre_mvn : bool (default: False)
    perform mean-variance normalization before stacking,
    then, feeding data to network.
  sad_name : {str, None}
    if None, or `sad_name` not found, don't applying SAD to
    the input feature before BNF
  remove_non_speech : bool (default: True)
    if True, remove non-speech frames using given SAD
  batch_size : int
    batch size when feeding data to the network, suggest
    to have as much data as possible.

  Note
  ----
  Order of preprocessing for BNF:
   - delta and delta delta extraction (optional)
   - mean_var_norm based on SAD frames statistics.
   - Stacking the left and right context frames.
   - Applying SAD indices.
   - Mean-variance normalization
   => BNFExtractor

  """

  def _prepare_network(self, network):
    from odin.nnet import NNOp
    if not isinstance(network, NNOp):
      raise ValueError("`network` must be instance of odin.nnet.NNOp")
    self.network = network

  def _apply_dnn(self, X):
    return self.network(X)

  def __getstate__(self):
    from odin import nnet as N, backend as K
    if not self.network.is_initialized:
      self.network()
    K.initialize_all_variables()
    return (self._input_name, self._output_name,
            self.use_sad, self.batch_size, self.stack_context, self.pre_mvn,
            N.serialize(self.network, binary_output=True))

  def __setstate__(self, states):
    from odin import nnet as N
    (self._input_name, self._output_name,
     self.use_sad, self.batch_size, self.stack_context, self.pre_mvn,
     self.network) = states
    self.network = N.deserialize(self.network,
                                 force_restore_vars=False)

# ===========================================================================
# Pitch
# ===========================================================================
class PitchExtractor(Extractor):
  """
  Parameters
  ----------
  threshold : float, optional
      Voice/unvoiced threshold. Default is 0.3 (as suggested for SWIPE)
      Threshold >= 1.0 is suggested for RAPT
  algo: 'swipe', 'rapt', 'avg'
      swipe - A Saw-tooth Waveform Inspired Pitch Estimation.
      rapt - a robust algorithm for pitch tracking.
      avg - apply swipe and rapt at the same time, then take average.
      Default is 'swipe'

  """

  def __init__(self, frame_length, step_length=None,
               threshold=0.5, fmin=20, fmax=400,
               algo='swipe', f0=False, input_name=('raw', 'sr')):
    super(PitchExtractor, self).__init__(input_name=input_name)
    self.threshold = threshold
    self.fmin = int(fmin)
    self.fmax = int(fmax)
    self.algo = algo
    self.f0 = f0
    self.frame_length = frame_length
    self.step_length = step_length

  def _transform(self, y_sr):
    y, sr = [y_sr[name] for name in self.input_name]
    frame_length, step_length = _extract_frame_step_length(
        sr, self.frame_length, self.step_length)
    # ====== extract pitch ====== #
    pitch_freq = pitch_track(y=y, sr=sr, step_length=step_length, fmin=self.fmin,
        fmax=self.fmax, threshold=self.threshold, otype='pitch',
        algorithm=self.algo)
    pitch_freq = np.expand_dims(pitch_freq, axis=-1)
    if self.f0:
      f0_freq = pitch_track(y=y, sr=sr, step_length=step_length, fmin=self.fmin,
          fmax=self.fmax, threshold=self.threshold, otype='f0',
          algorithm=self.algo)
      f0_freq = np.expand_dims(f0_freq, axis=-1)
      return {'pitch': pitch_freq,
              'f0': f0_freq}
    return {'pitch': pitch_freq}

# ===========================================================================
# SAD
# ===========================================================================
def _numba_thresholding(energy,
                        energy_threshold, energy_mean_scale,
                        frame_context, proportion_threshold):
  """ Using this numba function if at least 5 time faster than
  python/numpy implementation """
  n_frames = len(energy)
  # ====== normalize to [0, 1] ====== #
  e_min = np.min(energy)
  e_max = np.max(energy)
  energy = (energy - e_min) / (e_max - e_min)
  # ====== scale the threshold ====== #
  if energy_mean_scale != 0:
    energy_threshold += energy_mean_scale * np.sum(energy) / n_frames
  # ====== thresholding ====== #
  sad = np.empty(shape=(n_frames,))
  for t in range(n_frames):
    num_count = 0
    den_count = 0
    for t2 in range(t - frame_context, t + frame_context + 1):
      if 0 <= t2 < n_frames:
        den_count += 1
        if energy[t2] > energy_threshold:
          num_count += 1
    if num_count >= den_count * proportion_threshold:
      sad[t] = 1
    else:
      sad[t] = 0
  return sad, energy_threshold

try:
  import numba as nb
  _numba_thresholding = nb.jit(nopython=True, nogil=True)(_numba_thresholding)
except ImportError as e:
  pass


class SADthreshold(Extractor):

  """ Compute voice-activity vector for a file: 1 if we judge the frame as
  voiced, 0 otherwise.  There are no continuity constraints.
  This method is a very simple energy-based method which only looks
  at the first coefficient of "input_features", which is assumed to
  be a log-energy or something similar.  A cutoff is set-- we use
  a formula of the general type: cutoff = 5.0 + 0.5 * (average log-energy
  in this file), and for each frame the decision is based on the
  proportion of frames in a context window around the current frame,
  which are above this cutoff.

  This method is geared toward speaker-id applications and is not suitable
  for automatic speech recognition (ASR) because it makes independent
  decisions for each frame without imposing any notion of continuity.

  This function is optimized by numba, hence, the performance is
  comparable to kaldi-C code

  Parameters
  ----------
  energy_threshold : float (default: 0.55)
    value from [0., 1.], constant term in energy threshold
    for MFCC0 for SAD, the higher the value, the more strict
    the SAD

  energy_mean_scale : float (default: 0.5)
    If this is set, to get the actual threshold we let m be the mean
    log-energy of the file, and use `s*m + vad-energy-threshold`

  frames_context : float (default: 2)
    Number of frames of context on each side of central frame,
    in window for which energy is monitored

  proportion_threshold : float (default: 0.12)
    Parameter controlling the proportion of frames within
    the window that need to have more energy than the
    threshold

  smooth_window : int (default: 5)
    smooth the transition between SAD windows, the higher the value
    the more continuity of the SAD

  Note
  ----
  This implementation is slightly different from kaldi implementation,
  we normalize the energy to [0, 1] and thresholding based on
  these values

  This algorithm could fail if there is significant amount
  of noise in the audio, then it treats all frames as non-speech

  Copyright
  ---------
  Daniel Povey, voice-activity-detection.cc, kaldi toolkit

  """

  def __init__(self, energy_threshold=0.55, energy_mean_scale=0.5,
               frame_context=2, proportion_threshold=0.12, smooth_window=5,
               input_name='energy', output_name='sad'):
    super(SADthreshold, self).__init__(input_name=str(input_name),
                                       output_name=str(output_name))
    self.energy_threshold = float(energy_threshold)
    self.energy_mean_scale = float(energy_mean_scale)
    self.proportion_threshold = float(proportion_threshold)
    self.frame_context = int(frame_context)
    self.smooth_window = int(smooth_window)
    # ====== validate all parameters ====== #
    assert self.energy_mean_scale > 0, \
    'energy_mean_scale > 0, given: %.2f' % self.energy_mean_scale
    assert self.frame_context >= 0, \
    'frame_context >= 0, given: %d' % self.frame_context
    assert 0. < self.proportion_threshold < 1., \
    '0 < proportion_threshold < 1, given: %.2f' % self.proportion_threshold

  def _transform(self, X):
    # ====== preprocess ====== #
    energy = X[self.input_name]
    if energy.ndim > 1:
      energy = np.squeeze(energy)
    assert energy.ndim == 1, "Only support 1-D energy"
    sad, energy_threshold = _numba_thresholding(
        energy.astype('float32'),
        self.energy_threshold, self.energy_mean_scale,
        self.frame_context, self.proportion_threshold)
    sad = sad.astype('uint8')
    # ====== smooth the sad ====== #
    if self.smooth_window > 0:
      # at least 2 voice frames
      threshold = (2. / self.smooth_window)
      sad = smooth(x=sad, win=self.smooth_window, window='flat') >= threshold
    # ====== return the sad ====== #
    return {self.output_name: sad,
            '%s_threshold' % self.output_name: energy_threshold}

class SADgmm(Extractor):
  """ GMM-based SAD extractor

  Note
  ----
  This method can completely fail for very noisy audio, or audio
  with very long silence
  """

  def __init__(self, nb_mixture=3, nb_train_it=24 + 1, smooth_window=3,
               input_name='energy', output_name='sad'):
    super(SADgmm, self).__init__(input_name=input_name, output_name=output_name)
    self.nb_mixture = int(nb_mixture)
    self.nb_train_it = int(nb_train_it)
    self.smooth_window = int(smooth_window)

  def _transform(self, feat):
    # ====== select features type ====== #
    features = feat[self.input_name]
    if features.ndim > 1:
      features = features.sum(axis=-1)
    # ====== calculate VAD ====== #
    sad, sad_threshold = vad_energy(log_energy=features.ravel(),
        distrib_nb=self.nb_mixture, nb_train_it=self.nb_train_it)
    if self.smooth_window > 0:
      # at least 2 voice frames
      threshold = (2. / self.smooth_window)
      sad = smooth(sad, win=self.smooth_window, window='flat') >= threshold
    # ====== vad is only 0 and 1 so 'uint8' is enough ====== #
    sad = sad.astype('uint8')
    return {self.output_name: sad,
            '%s_threshold' % self.output_name: float(sad_threshold)}

# ===========================================================================
# Normalization
# ===========================================================================
class RASTAfilter(Extractor):
  """ RASTAfilter

  Specialized "Relative Spectral Transform" applying for MFCCs
  and PLP

  RASTA is a separate technique that applies a band-pass filter
  to the energy in each frequency subband in order to smooth over
  short-term noise variations and to remove any constant offset
  resulting from static spectral coloration in the speech channel
  e.g. from a telephone line

  Parameters
  ----------
  RASTA : bool
    R.A.S.T.A filter
  sdc : int (default: 1)
      Lag size for delta feature computation for
      "Shifted Delta Coefficients", if `sdc` > 0, the
      shifted delta features will be append to MFCCs

  References
  ----------
  [PLP and RASTA](http://www.ee.columbia.edu/ln/rosa/matlab/rastamat/)

  """

  def __init__(self, rasta=True, sdc=1,
               input_name='mfcc', output_name=None):
    super(RASTAfilter, self).__init__(input_name=as_tuple(input_name, t=string_types),
                                      output_name=output_name)
    self.rasta = bool(rasta)
    self.sdc = int(sdc)

  def _transform(self, feat):
    new_feat = []
    for name in self.input_name:
      mfcc = feat[name]
      # apply RASTA
      if self.rasta:
        mfcc = rastafilt(mfcc)
      # apply SDC if required
      if self.sdc >= 1:
        n_ceps = mfcc.shape[-1]
        mfcc = np.hstack([
            mfcc,
            shifted_deltas(mfcc, N=n_ceps, d=self.sdc,
                           P=3, k=n_ceps) # k = 7
        ])
      # store new feature
      new_feat.append(mfcc.astype("float32"))
    return new_feat

class AcousticNorm(Extractor):
  """
  Parameters
  ----------
  mean_var_norm : bool (default: True)
    mean-variance normalization
  windowed_mean_var_norm : bool (default: False)
    perform standardization on small windows, very computaiton
    intensive.
  sad_name : {str, None} (default: None)
    feature name of SAD indices, and only using statistics from
    SAD indexed frames for normalization
  ignore_sad_error : bool
    if True, when length of SAD and feature mismatch, still perform
    normalization, otherwise raise `RuntimeError`.

  """

  def __init__(self, input_name, output_name=None,
               mean_var_norm=True, windowed_mean_var_norm=False,
               win_length=301, var_norm=True,
               sad_name=None, ignore_sad_error=True):
    # ====== check which features will be normalized ====== #
    self.sad_name = str(sad_name) if isinstance(sad_name, string_types) else None
    self.ignore_sad_error = bool(ignore_sad_error)
    super(AcousticNorm, self).__init__(input_name=as_tuple(input_name, t=string_types),
                                       output_name=output_name)
    # ====== configs ====== #
    self.mean_var_norm = bool(mean_var_norm)
    self.windowed_mean_var_norm = bool(windowed_mean_var_norm)
    self.var_norm = bool(var_norm)
    # ====== check win_length ====== #
    win_length = int(win_length)
    if win_length % 2 == 0:
      raise ValueError("win_length must be odd number")
    if win_length < 3:
      raise ValueError("win_length must >= 3")
    self.win_length = win_length

  def _transform(self, feat):
    # ====== check SAD indices ====== #
    sad = None
    if self.sad_name is not None:
      sad = feat[self.sad_name]
      if sad.dtype != np.bool:
        sad = sad.astype(np.bool)
    # ====== normalize ====== #
    feat_normalized = []
    # all `features` is [t, f] shape
    for name in self.input_name:
      X = feat[name]
      X_sad = sad
      if sad is not None and len(sad) != len(X):
        if not self.ignore_sad_error:
          raise RuntimeError("Features with name: '%s' have length %d, but "
                             "given SAD has length %d" % (name, len(X), len(sad)))
        else:
          X_sad = None
      # mean-variance normalization
      if self.mean_var_norm:
        X = mvn(X, varnorm=self.var_norm, indices=X_sad)
      # windowed normalization
      if self.windowed_mean_var_norm:
        X = wmvn(X, w=self.win_length, varnorm=False, indices=X_sad)
      # update new features
      feat_normalized.append(X)
    return feat_normalized

class Read3ColSAD(Extractor):
  """ Read3ColSAD simple helper for applying 3 col
  SAD (name, start-in-second, end-in-second) to extracted acoustic features

  Parameters
  ----------
  path_or_map : {str, Mapping}
      path to folder contain all SAD files
      if Mapping, it is a map from `ref_key` to SAD dictionary
  step_length : float
      step length to convert second to frame index.
  ref_feat : str
      name of a reference features that must have the
      same length with the SAD.
  input_name : str (default: path)
      reference key in the pipeline can be use to get the coordinate value
      of SAD from given dictionary.
  file_regex: str
      regular expression for filtering the files name

  Return
  ------
   - add 'sad': array of SAD indexing

  """

  def __init__(self, path_or_map, step_length, ref_feat,
               input_name='path', output_name='sad',
               file_regex='.*'):
    super(Read3ColSAD, self).__init__(input_name=str(input_name))
    self.step_length = float(step_length)
    self.ref_feat = str(ref_feat)
    # ====== file regex ====== #
    file_regex = re.compile(str(file_regex))
    # ====== parse all file ====== #
    # read the SAD from file
    if is_string(path_or_map):
      sad = defaultdict(list)
      for f in get_all_files(path_or_map):
        if file_regex.search(os.path.basename(f)) is not None:
          with open(f, 'r') as f:
            for line in f:
              name, start, end = line.strip().split(' ')
              sad[name].append((float(start), float(end)))
    elif isinstance(path_or_map, Mapping):
      sad = path_or_map
    else:
      raise ValueError("`path` must be path to folder, or dictionary.")
    self._sad = sad

  def _transform(self, feat):
    # ====== get ref name ====== #
    name = feat[self.input_name]
    ref_feat = feat[self.ref_feat]
    n_samples = len(ref_feat)
    # ====== convert step_length ====== #
    step_length = self.step_length
    if step_length >= 1: # step_length is number of frames
      # now step_length is in second
      step_length = step_length / feat['sr']
    # ====== found SAD ====== #
    sad_indices = np.zeros(shape=(n_samples,), dtype=np.uint8)
    if name in self.sad and len(self.sad[name]) > 0:
      for start_sec, end_sec in self.sad[name]:
        start_idx = int(start_sec / step_length)
        end_idx = int(end_sec / step_length)
        # ignore zero SAD
        if end_idx - start_idx == 0:
          continue
        sad_indices[start_idx:end_idx] = 1
    # ====== return ====== #
    return {self.output_name: sad_indices}

class ApplyingSAD(Extractor):
  """ Applying SAD index to given features
  This extractor cutting voiced segments out, using extracted
  SAD labels previously


  Parameters
  ----------
  sad_name: str
      specific feature will be used name for the cutting
  threshold: None or float
      if `sad`, is continuous value, threshold need to be applied
  smooth_window: int (> 0)
      amount of adjacent frames will be taken into the SAD
  keep_unvoiced: bool
      if True, keep the whole audio file even though no SAD found
  stack_context : dict
      a dictionary mapping from feature name to number of
      context frames (a scalar for both left and right context)
      NOTE: the final frame length is: `context * 2 + 1`
  feat_name: str, or list of str
      all features' name will be applied.

  """

  def __init__(self, input_name, output_name=None, sad_name='sad',
               threshold=None, smooth_window=None, keep_unvoiced=False):
    super(ApplyingSAD, self).__init__(input_name=as_tuple(input_name, t=string_types),
                                      output_name=output_name)
    self.sad_name = str(sad_name)
    self.threshold = float(threshold) if is_number(threshold) else None
    self.smooth_window = int(smooth_window) if is_number(smooth_window) else None
    self.keep_unvoiced = bool(keep_unvoiced)

  def _transform(self, X):
    # ====== threshold sad to index ====== #
    sad = X[self.sad_name]
    if is_number(self.threshold):
      sad = (sad >= self.threshold).astype('int32')
    # ====== keep unvoiced or not ====== #
    if np.isclose(np.sum(sad), 0.):
      if not self.keep_unvoiced:
        return None
      else: # take all frames
        sad[:] = 1
    # ====== preprocessing sad ====== #
    if is_number(self.smooth_window) and self.smooth_window > 0:
      sad = smooth(sad, win=self.smooth_window, window='flat') > 0.
    sad = sad.astype('bool')
    # ====== start ====== #
    X_new = []
    for name in self.input_name:
      X_feat = X[name]
      assert len(sad) == max(X_feat.shape),\
          "Feature with name: %s, length of sad labels is: %d, but number of sample is: %s" % \
          (name, len(sad), max(X_feat.shape))
      # update feature
      X_new.append(X_feat[sad])
    return X_new



from __future__ import absolute_import, division, print_function

# Dependency imports
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
from six import string_types
from tensorflow.python.keras.utils import tf_utils as keras_tf_utils
# By importing `distributions` as `tfd`, docstrings will show
# `tfd.Distribution`. We import `bijectors` the same way, for consistency.
from tensorflow_probability.python import bijectors as tfb
from tensorflow_probability.python import distributions as tfd
from tensorflow_probability.python import layers as tfl
from tensorflow_probability.python.internal import \
    distribution_util as dist_util
from tensorflow_probability.python.layers.distribution_layer import _event_size
from tensorflow_probability.python.layers.internal import \
    distribution_tensor_coercible as dtc

from odin.bay.distributions import NegativeBinomialDisp, ZeroInflated

__all__ = [
    'DistributionLambda', 'MultivariateNormalLayer', 'BernoulliLayer',
    'DeterministicLayer', 'VectorDeterministicLayer', 'OneHotCategoricalLayer',
    'GammaLayer', 'DirichletLayer', 'GaussianLayer', 'NormalLayer',
    'LogNormalLayer', 'LogisticLayer', 'ZIBernoulliLayer',
    'update_convert_to_tensor_fn'
]

DistributionLambda = tfl.DistributionLambda
BernoulliLayer = tfl.IndependentBernoulli
LogisticLayer = tfl.IndependentLogistic


# ===========================================================================
# Helper
# ===========================================================================
def update_convert_to_tensor_fn(dist, fn):
  assert isinstance(dist, dtc._TensorCoercible), \
  "dist must be output from tfd.DistributionLambda"
  assert callable(fn), "fn must be callable"
  if isinstance(fn, property):
    fn = fn.fget
  dist._concrete_value = None
  dist._convert_to_tensor_fn = fn
  return dist


def _preprocess_eventshape(params, event_shape, n_dims=1):
  if isinstance(event_shape, string_types):
    if event_shape.lower().strip() == 'auto':
      event_shape = params.shape[-n_dims:]
    else:
      raise ValueError("Not support for event_shape='%s'" % event_shape)
  return event_shape


# ===========================================================================
# Simple distribution
# ===========================================================================
class DeterministicLayer(DistributionLambda):
  """
  ```none
  pmf(x; loc) = 1, if x == loc, else 0
  cdf(x; loc) = 1, if x >= loc, else 0
  ```
  """

  def __init__(self,
               event_shape=(),
               convert_to_tensor_fn=tfd.Distribution.sample,
               activity_regularizer=None,
               validate_args=False,
               **kwargs):
    super(DeterministicLayer,
          self).__init__(lambda t: type(self).new(t, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params, validate_args=False, name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'DeterministicLayer', [params]):
      params = tf.convert_to_tensor(value=params, name='params')
      return tfd.Deterministic(loc=params, validate_args=validate_args)

  @staticmethod
  def params_size(event_size, name=None):
    """ The number of `params` needed to create a single distribution. """
    return event_size


class VectorDeterministicLayer(DistributionLambda):
  """
  ```none
  pmf(x; loc)
    = 1, if All[Abs(x - loc) <= atol + rtol * Abs(loc)],
    = 0, otherwise.
  ```
  """

  def __init__(self,
               event_shape=(),
               convert_to_tensor_fn=tfd.Distribution.sample,
               activity_regularizer=None,
               validate_args=False,
               **kwargs):
    super(VectorDeterministicLayer,
          self).__init__(lambda t: type(self).new(t, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params, validate_args=False, name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'VectorDeterministicLayer', [params]):
      params = tf.convert_to_tensor(value=params, name='params')
      return tfd.VectorDeterministic(loc=params, validate_args=validate_args)

  @staticmethod
  def params_size(event_size, name=None):
    """ The number of `params` needed to create a single distribution. """
    return event_size


class OneHotCategoricalLayer(DistributionLambda):
  """ A `d`-variate OneHotCategorical Keras layer from `d` params.

  Parameters
  ----------
  convert_to_tensor_fn: callable
    that takes a `tfd.Distribution` instance and returns a
    `tf.Tensor`-like object. For examples, see `class` docstring.
    Default value: `tfd.Distribution.sample`.

  sample_dtype: `dtype`
    Type of samples produced by this distribution.
    Default value: `None` (i.e., previous layer's `dtype`).

  validate_args: `bool` (default `False`)
    When `True` distribution parameters are checked for validity
    despite possibly degrading runtime performance.
    When `False` invalid inputs may silently render incorrect outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  Note
  ----
  If input as probability values is given, it will be clipped by value
  [1e-8, 1 - 1e-8]

  """

  def __init__(self,
               event_shape=(),
               convert_to_tensor_fn=tfd.Distribution.sample,
               probs_input=False,
               sample_dtype=None,
               activity_regularizer=None,
               validate_args=False,
               **kwargs):
    super(OneHotCategoricalLayer, self).__init__(
        lambda t: type(self).new(t, probs_input, sample_dtype, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params, probs_input=False, dtype=None, validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'OneHotCategoricalLayer', [params]):
      params = tf.convert_to_tensor(value=params, name='params')
      return tfd.OneHotCategorical(
          logits=params if not probs_input else None,
          probs=tf.clip_by_value(params, 1e-8, 1 -
                                 1e-8) if probs_input else None,
          dtype=dtype or params.dtype.base_dtype,
          validate_args=validate_args)

  @staticmethod
  def params_size(event_size, name=None):
    """The number of `params` needed to create a single distribution."""
    return event_size


class DirichletLayer(DistributionLambda):
  """
  Parameters
  ----------
  pre_softplus : bool (default: False)
    applying softplus activation on the parameters before parameterizing

  clip_for_stable : bool (default: True)
    clipping the concentration into range [1e-3, 1e3] for stability

  """

  def __init__(self,
               event_shape='auto',
               pre_softplus=False,
               clip_for_stable=True,
               convert_to_tensor_fn=tfd.Distribution.sample,
               activity_regularizer=None,
               validate_args=False,
               **kwargs):
    super(DirichletLayer,
          self).__init__(lambda t: type(self).new(
              t, event_shape, pre_softplus, clip_for_stable, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params,
          event_shape='auto',
          pre_softplus=False,
          clip_for_stable=True,
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    event_shape = _preprocess_eventshape(params, event_shape)
    with tf.compat.v1.name_scope(name, 'DirichletLayer', [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      # Clips the Dirichlet parameters to the numerically stable KL region
      if pre_softplus:
        params = tf.nn.softplus(params)
      if clip_for_stable:
        params = tf.clip_by_value(params, 1e-3, 1e3)
      return tfd.Independent(
          tfd.Dirichlet(concentration=tf.reshape(params, output_shape),
                        validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'Dirichlet_params_size', [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return _event_size(event_shape, name=name or 'Dirichlet_params_size')


class GaussianLayer(DistributionLambda):
  """An independent normal Keras layer.

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.

  softplus_scale : bool
    if True, `scale = softplus(params) + softplus_inverse(1.0)`

  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.

  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               softplus_scale=True,
               convert_to_tensor_fn=tfd.Distribution.sample,
               activity_regularizer=None,
               validate_args=False,
               **kwargs):
    super(GaussianLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, softplus_scale, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          softplus_scale=True,
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'NormalLayer', [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      loc_params, scale_params = tf.split(params, 2, axis=-1)
      if softplus_scale:
        scale_params = tf.math.softplus(
            scale_params) + tfp.math.softplus_inverse(1.0)
      return tfd.Independent(
          tfd.Normal(loc=tf.reshape(loc_params, output_shape),
                     scale=tf.reshape(scale_params, output_shape),
                     validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'Normal_params_size', [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape, name=name or 'Normal_params_size')


class LogNormalLayer(DistributionLambda):
  """An independent LogNormal Keras layer.

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.

  softplus_scale : bool
    if True, `scale = softplus(params) + softplus_inverse(1.0)`

  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.

  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               softplus_scale=True,
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(LogNormalLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, softplus_scale, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          softplus_scale=True,
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'LogNormalLayer', [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      loc_params, scale_params = tf.split(params, 2, axis=-1)
      if softplus_scale:
        scale_params = tf.math.softplus(
            scale_params) + tfp.math.softplus_inverse(1.0)
      return tfd.Independent(
          tfd.LogNormal(loc=tf.reshape(loc_params, output_shape),
                        scale=tf.reshape(scale_params, output_shape),
                        validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'LogNormal_params_size', [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape, name=name or 'LogNormal_params_size')


class GammaLayer(DistributionLambda):
  """An independent Gamma Keras layer.

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.

  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.

  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(GammaLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params, event_shape=(), validate_args=False, name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'GammaLayer', [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      concentration_params, rate_params = tf.split(params, 2, axis=-1)
      return tfd.Independent(
          tfd.Gamma(concentration=tf.reshape(concentration_params,
                                             output_shape),
                    rate=tf.reshape(rate_params, output_shape),
                    validate_args=validate_args),
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'Gamma_params_size', [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape, name=name or 'Gamma_params_size')


# ===========================================================================
# Multivariate distribution
# ===========================================================================
class MultivariateNormalLayer(DistributionLambda):
  """A `d`-variate Multivariate Normal distribution Keras layer:

  Different covariance mode:
   - tril (lower triangle): `d + d * (d + 1) // 2` params.
   - diag (diagonal) : `d + d` params.
   - full (full) : `d + d * d` params.

  Typical choices for `convert_to_tensor_fn` include:

  - `tfd.Distribution.sample`
  - `tfd.Distribution.mean`
  - `tfd.Distribution.mode`
  - `lambda s: s.mean() + 0.1 * s.stddev()`

    Parameters
    ----------
    event_size: Scalar `int` representing the size of single draw from this
      distribution.

    covariance_type : {'diag', 'tril', 'full'}

    softplus_scale : bool
      if True, `scale = softplus(params) + softplus_inverse(1.0)`

    convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
      instance and returns a `tf.Tensor`-like object. For examples, see
      `class` docstring.
      Default value: `tfd.Distribution.sample`.

    validate_args: Python `bool`, default `False`. When `True` distribution
      parameters are checked for validity despite possibly degrading runtime
      performance. When `False` invalid inputs may silently render incorrect
      outputs.
      Default value: `False`.

    **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.
  """

  def __init__(self,
               event_size,
               covariance_type='diag',
               softplus_scale=True,
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(MultivariateNormalLayer,
          self).__init__(lambda t: type(self).new(
              t, event_size, covariance_type, softplus_scale, validate_args),
                         convert_to_tensor_fn,
                         activity_regularizer=activity_regularizer,
                         **kwargs)

  @staticmethod
  def new(params,
          event_size,
          covariance_type,
          softplus_scale,
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    covariance_type = str(covariance_type).lower().strip()
    assert covariance_type in ('full', 'tril', 'diag'), \
    "No support for given covariance_type: '%s'" % covariance_type

    if bool(softplus_scale):
      scale_fn = lambda x: tf.math.softplus(x) + tfp.math.softplus_inverse(1.0)
    else:
      scale_fn = lambda x: x

    with tf.compat.v1.name_scope(name, 'MultivariateNormalLayer',
                                 [params, event_size]):
      params = tf.convert_to_tensor(value=params, name='params')

      if covariance_type == 'tril':
        scale_tril = tfb.ScaleTriL(diag_shift=np.array(
            1e-5, params.dtype.as_numpy_dtype()),
                                   validate_args=validate_args)
        return tfd.MultivariateNormalTriL(
            loc=params[..., :event_size],
            scale_tril=scale_tril(scale_fn(params[..., event_size:])),
            validate_args=validate_args)

      elif covariance_type == 'diag':
        return tfd.MultivariateNormalDiag(loc=params[..., :event_size],
                                          scale_diag=scale_fn(
                                              params[..., event_size:]),
                                          validate_args=validate_args)

      elif covariance_type == 'full':
        return tfd.MultivariateNormalFullCovariance(
            loc=params[..., :event_size],
            covariance_matrix=tf.reshape(scale_fn(params[..., event_size:]),
                                         (event_size, event_size)),
            validate_args=validate_args)

  @staticmethod
  def params_size(event_size, covariance_type='diag', name=None):
    """The number of `params` needed to create a single distribution."""
    covariance_type = str(covariance_type).lower().strip()
    assert covariance_type in ('full', 'tril', 'diag'), \
    "No support for given covariance_type: '%s'" % covariance_type

    with tf.compat.v1.name_scope(name, 'MultivariateNormal_params_size',
                                 [event_size]):
      if covariance_type == 'tril':
        return event_size + event_size * (event_size + 1) // 2
      elif covariance_type == 'diag':
        return event_size + event_size
      elif covariance_type == 'full':
        return event_size + event_size * event_size


class ZIBernoulliLayer(DistributionLambda):
  """A Independent zero-inflated bernoulli keras layer

  Parameters
  ----------
  event_shape: integer vector `Tensor` representing the shape of single
    draw from this distribution.

  given_log_count : boolean
    is the input representing log count values or the count itself

  convert_to_tensor_fn: Python `callable` that takes a `tfd.Distribution`
    instance and returns a `tf.Tensor`-like object.
    Default value: `tfd.Distribution.sample`.

  validate_args: Python `bool`, default `False`. When `True` distribution
    parameters are checked for validity despite possibly degrading runtime
    performance. When `False` invalid inputs may silently render incorrect
    outputs.
    Default value: `False`.

  **kwargs: Additional keyword arguments passed to `tf.keras.Layer`.

  """

  def __init__(self,
               event_shape=(),
               given_logits=True,
               convert_to_tensor_fn=tfd.Distribution.sample,
               validate_args=False,
               activity_regularizer=None,
               **kwargs):
    super(ZIBernoulliLayer, self).__init__(
        lambda t: type(self).new(t, event_shape, given_logits, validate_args),
        convert_to_tensor_fn,
        activity_regularizer=activity_regularizer,
        **kwargs)

  @staticmethod
  def new(params,
          event_shape=(),
          given_logits=True,
          validate_args=False,
          name=None):
    """Create the distribution instance from a `params` vector."""
    with tf.compat.v1.name_scope(name, 'ZIBernoulliLayer',
                                 [params, event_shape]):
      params = tf.convert_to_tensor(value=params, name='params')
      event_shape = dist_util.expand_to_vector(tf.convert_to_tensor(
          value=event_shape, name='event_shape', dtype=tf.int32),
                                               tensor_name='event_shape')
      output_shape = tf.concat([
          tf.shape(input=params)[:-1],
          event_shape,
      ],
                               axis=0)
      (bernoulli_params, rate_params) = tf.split(params, 2, axis=-1)
      bernoulli_params = tf.reshape(bernoulli_params, output_shape)
      bern = tfd.Bernoulli(logits=bernoulli_params if given_logits else None,
                           probs=bernoulli_params if not given_logits else None,
                           validate_args=validate_args)
      zibern = ZeroInflated(count_distribution=bern,
                            logits=tf.reshape(rate_params, output_shape),
                            validate_args=validate_args)
      return tfd.Independent(
          zibern,
          reinterpreted_batch_ndims=tf.size(input=event_shape),
          validate_args=validate_args)

  @staticmethod
  def params_size(event_shape=(), name=None):
    """The number of `params` needed to create a single distribution."""
    with tf.compat.v1.name_scope(name, 'ZeroInflatedBernoulli_params_size',
                                 [event_shape]):
      event_shape = tf.convert_to_tensor(value=event_shape,
                                         name='event_shape',
                                         dtype=tf.int32)
      return 2 * _event_size(event_shape,
                             name=name or 'ZeroInflatedBernoulli_params_size')


# ===========================================================================
# Shortcut
# ===========================================================================
NormalLayer = GaussianLayer



from odin.preprocessing.base import make_pipeline, set_extractor_debug, Pipeline
from odin.preprocessing.processor import validate_features, FeatureProcessor, calculate_pca
from odin.preprocessing import base
from odin.preprocessing import speech
from odin.preprocessing import textgrid
from odin.preprocessing import signal
from odin.preprocessing import sequence
# from odin.preprocessing import image
# from odin.preprocessing import video
# from odin.preprocessing import text



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
# Copyright (c) 2017 - 2018 scVAE authors
# Copyright (c) 2018 - 2019 SISUA authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# TODO: NotImplementedError: <class 'odin.bay.distributions.zero_inflated.ZeroInflated'> does not support batch slicing; must implement _params_event_ndims.
# ======================================================================== #
"""The ZeroInflated distribution class."""
from __future__ import absolute_import, division, print_function

import tensorflow as tf
from tensorflow_probability.python.distributions import (Bernoulli, Independent,
                                                         distribution)
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.python.util.seed_stream import SeedStream

__all__ = ['ZeroInflated']


def _broadcast_rate(probs, *others):
  # make the shape broadcast-able
  others = list(others)
  others_ndims = [o.shape.ndims for o in others]
  assert len(set(others_ndims)) == 1
  others_ndims = others_ndims[0]

  probs_ndims = probs.shape.ndims

  if others_ndims < probs_ndims:
    for i in range(probs_ndims - others_ndims):
      others = [tf.expand_dims(o, -1) for o in others]
  elif others_ndims > probs_ndims:
    for i in range(others_ndims - probs_ndims):
      probs = tf.expand_dims(probs, -1)
  return [probs] + others


class ZeroInflated(distribution.Distribution):
  """zero-inflated distribution.

  The `zero-inflated` object implements batched zero-inflated distributions.
  The zero-inflated model is defined by a zero-inflation rate
  and a python list of `Distribution` objects.

  Methods supported include `log_prob`, `prob`, `mean`, `sample`, and
  `entropy_lower_bound`.
  """

  def __init__(self,
               count_distribution,
               inflated_distribution=None,
               logits=None,
               probs=None,
               validate_args=False,
               allow_nan_stats=True,
               name="ZeroInflated"):
    """Initialize a zero-inflated distribution.

    A `ZeroInflated` is defined by a zero-inflation rate (`inflated_distribution`,
    representing the probabilities of excess zeros) and a `Distribution` object
    having matching dtype, batch shape, event shape, and continuity
    properties (the dist).

    Parameters
    ----------
    count_distribution : A `tfp.distributions.Distribution` instance.
      The instance must have `batch_shape` matching the zero-inflation
      distribution.

    inflated_distribution: `tfp.distributions.Bernoulli`-like instance.
      Manages the probability of excess zeros, the zero-inflated rate.
      Must have either scalar `batch_shape` or `batch_shape` matching
      `count_distribution.batch_shape`.

    logits: An N-D `Tensor` representing the log-odds of a excess zeros
      A zero-inflation rate, where the probability of excess zeros is
      sigmoid(logits).
      Only one of `logits` or `probs` should be passed in.

    probs: An N-D `Tensor` representing the probability of a zero event.
      Each entry in the `Tensor` parameterizes an independent
      ZeroInflated distribution.
      Only one of `logits` or `probs` should be passed in.

    validate_args: Python `bool`, default `False`. If `True`, raise a runtime
      error if batch or event ranks are inconsistent between pi and any of
      the distributions. This is only checked if the ranks cannot be
      determined statically at graph construction time.

    allow_nan_stats: Boolean, default `True`. If `False`, raise an
     exception if a statistic (e.g. mean/mode/etc...) is undefined for any
      batch member. If `True`, batch members with valid parameters leading to
      undefined statistics will return NaN for this statistic.

    name: A name for this distribution (optional).

    References
    ----------
    Liu, L. & Blei, D.M.. (2017). Zero-Inflated Exponential Family Embeddings.
    Proceedings of the 34th International Conference on Machine Learning,
    in PMLR 70:2140-2148

    """
    parameters = dict(locals())
    self._runtime_assertions = []

    with tf.compat.v1.name_scope(name) as name:
      if not isinstance(count_distribution, distribution.Distribution):
        raise TypeError("count_distribution must be a Distribution instance"
                        " but saw: %s" % count_distribution)
      self._count_distribution = count_distribution

      if inflated_distribution is None:
        inflated_distribution = Bernoulli(logits=logits,
                                          probs=probs,
                                          dtype=tf.int32,
                                          validate_args=validate_args,
                                          allow_nan_stats=allow_nan_stats,
                                          name="ZeroInflatedRate")
      elif not isinstance(inflated_distribution, distribution.Distribution):
        raise TypeError("inflated_distribution must be a Distribution instance"
                        " but saw: %s" % inflated_distribution)
      self._inflated_distribution = inflated_distribution

      if self._count_distribution.batch_shape.ndims is None:
        raise ValueError(
            "Expected to know rank(batch_shape) from count_disttribution")
      if self._inflated_distribution.batch_shape.ndims is None:
        raise ValueError(
            "Expected to know rank(batch_shape) from inflated_distribution")

      # create Independent Bernoulli distribution that the batch_shape
      # of count_distribution matching batch_shape of inflated_distribution
      inflated_batch_ndims = self._inflated_distribution.batch_shape.ndims
      count_batch_ndims = self._count_distribution.batch_shape.ndims
      if count_batch_ndims < inflated_batch_ndims:
        self._inflated_distribution = Independent(
            self._inflated_distribution,
            reinterpreted_batch_ndims=inflated_batch_ndims - count_batch_ndims,
            name="ZeroInflatedRate")
      elif count_batch_ndims > inflated_batch_ndims:
        raise ValueError(
            "count_distribution has %d-D batch_shape, which smaller"
            "than %d-D batch_shape of inflated_distribution" %
            (count_batch_ndims, inflated_batch_ndims))

      # Ensure that all batch and event ndims are consistent.
      if validate_args:
        self._runtime_assertions.append(
            tf.assert_equal(
                self._count_distribution.batch_shape_tensor(),
                self._inflated_distribution.batch_shape_tensor(),
                message=(
                    "dist batch shape must match logits|probs batch shape")))

    # We let the zero-inflated distribution access _graph_parents since its arguably
    # more like a baseclass.
    reparameterization_type = [
        self._count_distribution.reparameterization_type,
        self._inflated_distribution.reparameterization_type
    ]
    if any(i == reparameterization.NOT_REPARAMETERIZED
           for i in reparameterization_type):
      reparameterization_type = reparameterization.NOT_REPARAMETERIZED
    else:
      reparameterization_type = reparameterization.FULLY_REPARAMETERIZED

    super(ZeroInflated,
          self).__init__(dtype=self._count_distribution.dtype,
                         reparameterization_type=reparameterization_type,
                         validate_args=validate_args,
                         allow_nan_stats=allow_nan_stats,
                         parameters=parameters,
                         graph_parents=self._count_distribution._graph_parents +
                         self._inflated_distribution._graph_parents,
                         name=name)

  @property
  def logits(self):
    """Log-odds of a `1` outcome (vs `0`)."""
    if isinstance(self._inflated_distribution, Independent):
      return self._inflated_distribution.distribution.logits_parameter()
    return self._inflated_distribution.logits_parameter()

  @property
  def probs(self):
    """Probability of a `1` outcome (vs `0`)."""
    if isinstance(self._inflated_distribution, Independent):
      return self._inflated_distribution.distribution.probs_parameter()
    return self._inflated_distribution.probs_parameter()

  @property
  def count_distribution(self):
    return self._count_distribution

  @property
  def inflated_distribution(self):
    return self._inflated_distribution

  def _batch_shape_tensor(self):
    return self._count_distribution._batch_shape_tensor()

  def _batch_shape(self):
    return self._count_distribution._batch_shape()

  def _event_shape_tensor(self):
    return self._count_distribution._event_shape_tensor()

  def _event_shape(self):
    return self._count_distribution._event_shape()

  def _mean(self):
    with tf.compat.v1.control_dependencies(self._runtime_assertions):
      # These should all be the same shape by virtue of matching
      # batch_shape and event_shape.
      probs, d_mean = _broadcast_rate(self.probs,
                                      self._count_distribution.mean())
      return (1 - probs) * d_mean

  def _variance(self):
    """
    (1 - pi) * (d.var + d.mean^2) - [(1 - pi) * d.mean]^2

    Note: mean(ZeroInflated) = (1 - pi) * d.mean
    where:
     - pi is zero-inflated rate
     - d is count distribution
    """
    with tf.compat.v1.control_dependencies(self._runtime_assertions):
      # These should all be the same shape by virtue of matching
      # batch_shape and event_shape.
      d = self._count_distribution

      probs, d_mean, d_variance = _broadcast_rate(self.probs, d.mean(),
                                                  d.variance())
      return (1 - probs) * \
      (d_variance + tf.square(d_mean)) - \
      tf.math.square(self._mean())

  def _log_prob(self, x):
    with tf.compat.v1.control_dependencies(self._runtime_assertions):
      eps = tf.cast(1e-8, x.dtype)
      x = tf.convert_to_tensor(x, name="x")
      d = self._count_distribution
      pi = self.probs

      log_prob = d.log_prob(x)
      prob = tf.math.exp(log_prob)

      # make pi and anything come out of count_distribution
      # broadcast-able
      pi, prob, log_prob = _broadcast_rate(pi, prob, log_prob)

      # This equation is validated
      # Equation (13) reference: u_{ij} = 1 - pi_{ij}
      y_0 = tf.math.log(pi + (1 - pi) * prob)
      y_1 = tf.math.log(1 - pi) + log_prob
      return tf.where(x <= eps, y_0, y_1)

  def _prob(self, x):
    return tf.math.exp(self._log_prob(x))

  def _sample_n(self, n, seed):
    with tf.compat.v1.control_dependencies(self._runtime_assertions):
      seed = SeedStream(seed, salt="ZeroInflated")
      mask = self.inflated_distribution.sample(n, seed())
      samples = self.count_distribution.sample(n, seed())
      mask, samples = _broadcast_rate(mask, samples)
      # mask = 1 => new_sample = 0
      # mask = 0 => new_sample = sample
      return samples * tf.cast(1 - mask, samples.dtype)

  # ******************** shortcut for denoising ******************** #
  def denoised_mean(self):
    return self.count_distribution.mean()

  def denoised_variance(self):
    return self.count_distribution.variance()
